from __future__ import annotations

import math

from AoE2ScenarioParser.datasets.dataset_enum import _DataSetIntEnums, _DataSetIntFlags


class DiplomacyState(_DataSetIntEnums):
    """
    This enum class provides the integer values used to reference the diplomacy states in the game. Used in the 'Change
    Diplomacy' effect and the 'Diplomacy State' condition

    **Examples**

    >>> DiplomacyState.ALLY
    <DiplomacyState.ALLY: 0>
    """
    ALLY = 0
    NEUTRAL = 1
    ENEMY = 3


class Operation(_DataSetIntEnums):
    """
    This enum class provides the integer values used to reference the operations in the game. Used in a lot of effects
    like 'Modify Attribute' to control whether an attribute is set, added to, multiplied or divided by a value.

    **Examples**

    >>> Operation.MULTIPLY
    <Operation.MULTIPLY: 4>
    """
    SET = 1
    ADD = 2
    SUBTRACT = 3
    MULTIPLY = 4
    DIVIDE = 5


class AttackStance(_DataSetIntEnums):
    """
    This enum class provides the integer values used to reference the different unit stances in the game. Used in the
    'Change Object Stance' effect

    **Examples**

    >>> AttackStance.AGGRESSIVE_STANCE
    <AttackStance.AGGRESSIVE_STANCE: 0>
    """
    AGGRESSIVE_STANCE = 0
    DEFENSIVE_STANCE = 1
    STAND_GROUND = 2
    NO_ATTACK_STANCE = 3


class UnitAIAction(_DataSetIntEnums):
    """
    This enum class provides the integer values used to reference the unit AI actions in the game. Used in the 'Object
    Has Action' condition.

    **Examples**

    >>> UnitAIAction.ATTACK
    <UnitAIAction.ATTACK: 1>
    """
    ANY = 0
    """Fires if the unit has any action"""
    ATTACK = 1
    """Fires when the unit is attacking any unit"""
    BUILD = 3
    """Unknown"""
    CONVERT = 5
    """Fires when a monk or missionary is converting any unit"""
    DEFEND = 2
    """Unknown"""
    ENTER = 18
    """Unknown"""
    EVADE = 17
    """Unknown"""
    EXPLORE = 6
    """Unknown"""
    FOLLOW = 13
    """Fires when a unit is following any unit"""
    GATHER = 10
    """Fires when a villager or fishing ship is gathering resources"""
    HEAL = 4
    """Fires when a monk or missionary is healing another unit"""
    HUNT = 14
    """
    Fires when a hunter is gathering from a corpse of a hunted animal. Does **NOT** fire when the hunter is chasing the
    animal
    """
    IDLE = 24
    """Fires when a unit is standing still"""
    MOVE = 11
    """Fires when a unit is moving"""
    PATROL = 12
    """Fires when a unit is patrolling"""
    RELIC = 23
    """Unknown"""
    REPAIR = 19
    """Fires when a villager is repairing any other unit"""
    RESEARCH = 21
    """Unknown"""
    RETREAT = 9
    """Unknown"""
    RUNAWAY = 8
    """Unknown"""
    STOP = 7
    """Unknown"""
    TRADE = 16
    """Fires when a trade cart is returning to any of its own player's markets"""
    TRAIN = 20
    """Unknown"""
    TRANSPORT = 15
    """Unknown"""
    UNLOAD = 22
    """Fires when a transport ship is tasked to unload objects. Note that the transport gets stuck in this state!"""


class ButtonLocation(_DataSetIntEnums):
    """
    This enum class provides the integer values used to reference the button locations in the game. These button
    locations are what determines where a unit's train button or a research's research button appears in a building's
    UI

    **Examples**

    >>> ButtonLocation.r2c2
    <ButtonLocation.r2c2: 7>
    >>> ButtonLocation.r3c1.attribute_presentation()
    'row_3_col_1'
    """
    _r1c1 = 0
    r1c1 = 1
    r1c2 = 2
    r1c3 = 3
    r1c4 = 4
    r1c5 = 5
    r2c1 = 6
    r2c2 = 7
    r2c3 = 8
    r2c4 = 9
    r2c5 = 10
    r3c1 = 11
    r3c2 = 12
    r3c3 = 13
    r3c4 = 14

    # r3c5 = 15  # Doesn't actually work in-game. Probably to make space for the arrow key.

    def attribute_presentation(self):
        val = self.value or 1  # Change VAL 0 to 1
        row = math.ceil(val / 5)
        col = val - (row - 1) * 5
        return f"row_{row}_col_{col}"

    @classmethod
    def row_col(cls, row: int, col: int) -> int:
        """
        Get the button location ID of the row, column specified

        Args:
            row: The number of the row starting from the top (1-5)
            col: The number of the column starting from the left (1-3)

        Returns:
            The button location ID of the (row, column) location specified
        """
        return cls((row - 1) * 5 + col)


class PanelLocation(_DataSetIntEnums):
    """
    This enum class provides the integer values used to reference the panel positons in the game. Used in the 'Display
    Information' effect.

    **Examples**

    >>> PanelLocation.TOP
    <PanelLocation.TOP: 0>
    """
    TOP = 0
    """Panel at the top of the screen. ~13% from the top"""
    BETWEEN = 1
    """Panel between the top and the center of the screen. ~33% from the top"""
    CENTER = 2
    """Panel close to the center of the screen. ~45% from the top"""


class TimeUnit(_DataSetIntEnums):
    """
    This enum class provides the integer values used to reference the unit of time used in an effect. Used in the
    'Display Timer' effect.

    **Examples**

    >>> TimeUnit.YEARS
    <TimeUnit.YEARS: 0>
    """
    YEARS = 0
    """In-Game years. A year is 5 seconds in-game time."""
    MINUTES = 1
    """In-Game minutes."""
    SECONDS = 2
    """In-Game seconds."""
    MINUTES_AND_SECONDS = 3
    """In-Game Minutes abd seconds."""


class VisibilityState(_DataSetIntEnums):
    """
    This enum class provides the integer values used to reference visibility state of a player for another player in the
    game. Used in the 'Set Visibility State' effect.

    **Examples**

    >>> VisibilityState.EXPLORED
    <VisibilityState.EXPLORED: 1>
    """
    VISIBLE = 0
    EXPLORED = 1
    INVISIBLE = 2


class DifficultyLevel(_DataSetIntEnums):
    """
    This enum class provides the integer values used to reference difficulty level of the game. Used in the 'Difficulty
    Level' condition.

    **Examples**

    >>> DifficultyLevel.HARD
    <DifficultyLevel.HARD: 1>
    """
    EASIEST = 4
    STANDARD = 3
    MODERATE = 2
    HARD = 1
    HARDEST = 0
    # EXTREME = 5  # ???


class TechnologyState(_DataSetIntEnums):
    """
    This enum class provides the integer values used to reference technology state of a technology in the game. Used in
    the 'Technology State' condition.

    **Examples**

    >>> TechnologyState.NOT_READY
    <TechnologyState.NOT_READY: 0>
    """
    DISABLED = -1
    NOT_READY = 0
    """A tech that is not available to be researched (Bombard Tower is not ready before chemistry is researched)"""
    READY = 1
    """A tech that is available to be researched (Bombard Tower is ready after chemistry is researched)"""
    RESEARCHING = 2
    """A tech that is currently being researched"""
    DONE = 3
    """A Tech that has already been researched"""
    QUEUED = 4
    """A tech that is waiting in queue to be researched"""


class Comparison(_DataSetIntEnums):
    """
    This enum class provides the integer values used to reference the comparisons in the game. Used in a lot of
    conditions like 'Accumulate Attribute' to perform logical operations on the attribute values

    **Examples**

    >>> Comparison.EQUAL
    <Comparison.EQUAL: 0>
    """
    EQUAL = 0
    LESS = 1
    LARGER = 2
    LESS_OR_EQUAL = 3
    LARGER_OR_EQUAL = 4


class ObjectAttribute(_DataSetIntEnums):
    """
    This enum class provides the integer values used to reference all the different object attributes in the game. Used
    in the 'Modify Attribute' effect to control which attribute of an object is modified.

    **Examples**

    >>> ObjectAttribute.LINE_OF_SIGHT
    <ObjectAttribute.LINE_OF_SIGHT: 1>
    """
    HIT_POINTS = 0
    LINE_OF_SIGHT = 1
    GARRISON_CAPACITY = 2
    UNIT_SIZE_X = 3
    UNIT_SIZE_Y = 4
    MOVEMENT_SPEED = 5
    ROTATION_SPEED = 6
    ARMOR = 8
    ATTACK = 9
    ATTACK_RELOAD_TIME = 10
    ACCURACY_PERCENT = 11
    MAX_RANGE = 12
    WORK_RATE = 13
    CARRY_CAPACITY = 14
    BASE_ARMOR = 15
    PROJECTILE_UNIT = 16
    BUILDING_ICON_OVERRIDE = 17
    TERRAIN_DEFENSE_BONUS = 18
    ENABLE_SMART_PROJECTILES = 19
    MINIMUM_RANGE = 20
    AMOUNT_OF_1ST_RESOURCES = 21
    BLAST_WIDTH = 22
    SEARCH_RADIUS = 23
    BONUS_DAMAGE_RESIST = 24
    ICON_ID = 25
    HERO_STATUS = 40
    FRAME_DELAY = 41
    TRAIN_LOCATION = 42
    TRAIN_BUTTON = 43
    BLAST_ATTACK_LEVEL = 44
    BLAST_DEFENSE_LEVEL = 45
    SHOWN_ATTACK = 46
    SHOWN_RANGE = 47
    SHOWN_MELEE_ARMOR = 48
    SHOWN_PIERCE_ARMOR = 49
    OBJECT_NAME_ID = 50
    SHORT_DESCRIPTION_ID = 51
    TERRAIN_RESTRICTION_ID = 53
    DEAD_UNIT_ID = 57
    HOTKEY_ID = 58
    RESOURCE_COSTS = 100
    TRAIN_TIME = 101
    TOTAL_MISSILES = 102
    FOOD_COSTS = 103
    WOOD_COSTS = 104
    GOLD_COSTS = 105
    STONE_COSTS = 106
    MAX_TOTAL_MISSILES = 107
    GARRISON_HEAL_RATE = 108
    """Hidden in the editor, but does work! Do not open effect in editor, will cause it to reset"""
    REGENERATION_RATE = 109


class Attribute(_DataSetIntEnums):
    """
    This enum class provides the integer values used to reference all the player resources in the game. Used in effects
    and conditions like 'Accumulate Attribute' and 'Modify Resource'

    **Examples**

    >>> Attribute.FOOD
    <Attribute.FOOD: 0>
    """

    FOOD = 0
    """
    Food amount of the source player
    """
    WOOD = 1
    """
    Wood amount of the source player
    """
    STONE = 2
    """
    Stone amount of the source player
    """
    GOLD = 3
    """
    Gold amount of the source player
    """
    POPULATION_CAP = 4
    """
    Current max pop of the source player
    """
    CONVERSION_RANGE = 5
    """
    Unknown amount of the source player
    """
    CURRENT_AGE = 6
    """
    Age name and icon at the top of the screen of the source player

    - Default Values:

        - 0:  Dark Age

        - 1:  Feudal Age

        - 2:  Castle Age

        - 3:  Imperial Age

    - Additional Information: Setting this to an amount higher than 3 cycles the icon but keeps the age at imperial
    """
    RELICS = 7
    """
    Number of relics held by the source player
    """
    TRADE_BONUS = 8
    """
    Unused amount of the source player
    """
    TRADE_GOODS = 9
    """
    Unused amount of the source player
    """
    TRADE_PRODUCTION = 10
    """
    Unused amount of the source player
    """
    CURRENT_POPULATION = 11
    """
    Current pop of the source player
    """
    CORPSE_DECAY_TIME = 12
    """
    Time taken by corpses to decay for the source player

    - Additional Information: Doesn't seem to do anything when changed
    """
    REMARKABLE_DISCOVERY = 13
    """
    Unknown amount of the source player
    """
    MONUMENTS_CAPTURED = 14
    """
    Number of monuments owned by the source player
    """
    MEAT_STORAGE = 15
    """
    Unknown amount of the source player
    """
    BERRY_STORAGE = 16
    """
    Unknown amount of the source player
    """
    FISH_STORAGE = 17
    """
    Unknown amount of the source player
    """
    UNUSED_RESOURCE_018 = 18
    """
    Unused amount of the source player
    """
    TOTAL_UNITS_OWNED = 19
    """
    Total units owned by the source player
    """
    KILLS = 20
    """
    Total units killed of the source player
    """
    RESEARCH_COUNT = 21
    """
    Research count of the source player
    """
    EXPLORATION = 22
    """
    Percent map explored by the source player
    """
    CASTLE_AGE_TECH_ID = 23
    """
    Always 102 

    - Additional Information: Nothing happens when you change this
    """
    IMPERIAL_AGE_TECH_ID = 24
    """
    Always 103 

    - Additional Information: Nothing happens when you change this
    """
    FEUDAL_AGE_TECH_ID = 25
    """
    Always 101 

    - Additional Information: Nothing happens when you change this
    """
    ATTACK_SOUND_EFFECT_ID = 26
    """
    Always 0 

    - Additional Information: Nothing happens when you change this
    """
    ENABLE_MONK_CONVERSION = 27
    """
    Boolean: allow enemy monk conversions by the source player

    - Default Values:

        - 0:  Don't Allow

        - $\geq$ 1:  Allow

    - Additional Information: Set to 1 when atonement is researched
    """
    ENABLE_BUILDING_CONVERSIONS = 28
    """
    Boolean: allow enemy building conversions by the source player.

    - Default Values:

        - 0:  Don't Allow

        - 1:  Allow

        - $\geq$2:  Monks can convert buildings from range

    - Additional Information: Set to 1 when redemption is researched
    """
    UNUSED_RESOURCE_029 = 29
    """
    Unused amount of the source player
    """
    BUILDING_LIMIT = 30
    """
    Unused amount of the source player
    """
    FOOD_LIMIT = 31
    """
    Unused amount of the source player
    """
    BONUS_POPULATION_CAP = 32
    """
    Additional max pop space of the source player

    - Additional Information: 10 for goths
    """
    FOOD_MAINTENANCE = 33
    """
    Unknown amount of the source player
    """
    FAITH = 34
    """
    Boolean: faith researched by the source player

    - Additional Information: Set to 1 when faith is researched, ONLY  a boolean value, does not force faith's effects if manually set to 1
    """
    FAITH_RECHARGE_RATE = 35
    """
    Monk faith recovery rate of the source player
    """
    FARM_FOOD = 36
    """
    Max farm food of the source player
    """
    CIVILIAN_POPULATION = 37
    """
    Civilian pop of the source player
    """
    UNUSED_RESOURCE_038 = 38
    """
    Unused amount of the source player
    """
    ALL_TECHS_ACHIEVED = 39
    """
    Boolean: researched all enabled techs by the source player

    - Default Values:

        - 0:  when all techs not researched

        - 1:  when all techs are researched
    """
    MILITARY_POPULATION = 40
    """
    Military pop of the source player
    """
    CONVERSIONS = 41
    """
    Number of units converted by the source player
    """
    WONDER = 42
    """
    Number of standing wonders of the source player
    """
    RAZINGS = 43
    """
    Number of buildings destroyed by the source player
    """
    KILL_RATIO = 44
    """
    Ceil of kills/deaths of the source player
    """
    PLAYER_KILLED = 45
    """
    Boolean: survival to finish of the source player

    - Default Values:

        - 0:  No

        - 1:  Yes
    """
    TRIBUTE_INEFFICIENCY = 46
    """
    Tribute tax fraction imposed on the source player
    """
    GOLD_MINING_PRODUCTIVITY = 47
    """
    Amount of gold mined multiplier of the source player

    - Default Values:

        - 1:  Generic

        - 1.15:  Mayans

    - Additional Information: Since this works by multiplying the amount of resources gathered by a villager, it has a side effect of increasing the gather rate as well. In the case of Mayans, This is compensated for by reducing villager work rate by 15%
    """
    TOWN_CENTER_UNAVAILABLE = 48
    """
    Boolean: allow building tcs for the source player

    - Default Values:

        - 0:  No (Sudden Death)

        - 1:  Yes
    """
    GOLD_COUNTER = 49
    """
    Total gold collected by the source player
    """
    REVEAL_ALLY = 50
    """
    Boolean: show ally los for the source player

    - Default Values:

        - 0:  No Allied Vision

        - 1:  Allied Vision

    - Additional Information: Note - Once enabled, Allied vison cannot be undone
    """
    HOUSES = 51
    """
    Unused amount of the source player
    """
    MONASTERIES = 52
    """
    Number of monasteries of the source player
    """
    TRIBUTE_SENT = 53
    """
    Amount tributed by the source player
    """
    ALL_MONUMENTS_CAPTURED = 54
    """
    Boolean: all monuments captured by the source player
    """
    ALL_RELICS_CAPTURED = 55
    """
    Boolean: all relics captured by the source player
    """
    ORE = 56
    """
    SWGB ore amount of the source player
    """
    CAPTURED_UNIT = 57
    """
    Number of units kidnapped by the source player
    """
    DARK_AGE_TECH_ID = 58
    """
    Always 104 

    - Additional Information: Nothing happens when you change this
    """
    TRADE_GOOD_QUALITY = 59
    """
    Unused amount of the source player
    """
    TRADE_MARKET_LEVEL = 60
    """
    Unused amount of the source player
    """
    FORMATIONS = 61
    """
    Unused amount of the source player
    """
    BUILDING_HOUSE_RATE = 62
    """
    Unknown amount of the source player
    """
    GATHER_TAX_RATE = 63
    """
    Unknown amount of the source player
    """
    GATHER_ACCUMULATION = 64
    """
    Unknown amount of the source player
    """
    SALVAGE_DECAY_RATE = 65
    """
    Boat corpse decay rate of the source player

    - Additional Information: Changing this doesn't seem to do anything
    """
    ALLOW_FORMATIONS = 66
    """
    Unused amount of the source player
    """
    CAN_CONVERT = 67
    """
    Boolean: allow enemy object conversions amount of the source player

    - Default Values:

        - 0:  Don't allow even regular unit conversion

        - 1:  Allows regular unit conversion
    """
    HITPOINTS_KILLED = 68
    """
    Cumulative hp of all units killed by the source player
    """
    PLAYER1_KILLS = 69
    """
    Number of player 1 units killed by the source player
    """
    PLAYER2_KILLS = 70
    """
    Number of player 2 units killed by the source player
    """
    PLAYER3_KILLS = 71
    """
    Number of player 3 units killed by the source player
    """
    PLAYER4_KILLS = 72
    """
    Number of player 4 units killed by the source player
    """
    PLAYER5_KILLS = 73
    """
    Number of player 5 units killed by the source player
    """
    PLAYER6_KILLS = 74
    """
    Number of player 6 units killed by the source player
    """
    PLAYER7_KILLS = 75
    """
    Number of player 7 units killed by the source player
    """
    PLAYER8_KILLS = 76
    """
    Number of player 8 units killed by the source player
    """
    CONVERSION_RESISTANCE = 77
    """
    Coefficient of conversion resistance of the source player

    - Additional Information: Probability of conversion is divided by this value for ALL source player units, Teuton team bonus for conversion resistance works by increasing this.
    """
    TRADE_VIG_RATE = 78
    """
    Market exchange rate fraction for the source player

    - Default Values:

        - 0.3:  Generic rate

        - 0.15:  after guilds

        - 0.05:  for saracens
    """
    STONE_MINING_PRODUCTIVITY = 79
    """
    Amount of stone mined multiplier of the source player

    - Default Values:

        - 1:  Generic

        - 1.15:  Mayans

    - Additional Information: Since this works by multiplying the amount of resources gathered by a villager, it has a side effect of increasing the gather rate as well. In the case of Mayans, This is compensated for by reducing villager work rate by 15%
    """
    QUEUED_COUNT = 80
    """
    Amount of units in queue of the source player

    - Additional Information: Note that only the units waiting to be trained are considered in the queue so if an archery range has 3 archers being made, there is 1 archer that is being trained and 2 archers that are in queue
    """
    TRAINING_COUNT = 81
    """
    Amount of units being trained of the source player

    - Additional Information: Note that only the FIRST unit in each building is considered as being trained so if a town centre has 4 villagers being made, there is 1 archer that is being trained and 3 villagers that are in queue
    """
    START_WITH_PACKED_TOWN_CENTRE = 82
    """
    Boolean: started with PTWC of the source player

    - Additional Information: Setting this to 1 in an RMS allows for starting with PTWC. Manually changing this in the editor does nothing
    """
    BOARDING_RECHARGE_RATE = 83
    """
    ABGAL faith recharge rate amount of the source player

    - Additional Information: This is similar to monk's faith except for a special ship unit 536 called the ABGAL in the editor that can convert ships form 1 range away
    """
    STARTING_VILLAGERS = 84
    """
    Number of starting villagers of the source player

    - Default Values:

        - 3:  For generic civs

        - 4:  For Mayans

        - 6:  For Chinese

    - Additional Information: Only works for RMS, Changing this manually in the editor does nothing x2 on budapest, x3 on metropolis
    """
    RESEARCH_COST_MODIFIER = 85
    """
    Reduce tech cost to fraction for the source player
    """
    RESEARCH_TIME_MODIFIER = 86
    """
    Reduce tech research time to fraction for the source player
    """
    CONVERT_BOATS = 87
    """
    Boolean: allow monks to convert boats amount of the source player
    """
    FISH_TRAP_FOOD = 88
    """
    Max fishtrap food of the source player
    """
    HEAL_RATE_MODIFIER = 89
    """
    Monk healing rate of the source player
    """
    HEAL_RANGE = 90
    """
    Monk heal range of the source player
    """
    STARTING_FOOD = 91
    """
    Starting food amount of the source player

    - Additional Information: Only works for RMS. Changing this manually in the editor does nothing but its a way to check starting food amount
    """
    STARTING_WOOD = 92
    """
    Starting wood amount of the source player

    - Additional Information: Only works for RMS. Changing this manually in the editor does nothing but its a way to check starting wood amount
    """
    STARTING_STONE = 93
    """
    Starting stone amount of the source player

    - Additional Information: Only works for RMS. Changing this manually in the editor does nothing but its a way to check starting stone amount
    """
    STARTING_GOLD = 94
    """
    Starting gold amount of the source player

    - Additional Information: Only works for RMS. Changing this manually in the editor does nothing but its a way to check starting gold amount
    """
    ENABLE_PTWC_OR_KIDNAP_OR_LOOT = 95
    """
    Enable town centre packing for the source player

    - Default Values:

        - 0:  Normal

        - 1:  Enables Pack Button on TC but it is bugged, if you click it then TC goes berserk

        - $\geq$2:  No noticeable effect
    """
    BERSERKER_HEAL_TIMER = 96
    """
    Time difference between berserker heal rate for source player

    - Additional Information: This is divided by two every time berserkergang is researched
    """
    DOMINANT_SHEEP_CONTROL = 97
    """
    Boolean: force sheep conversion of the source player

    - Default Values:

        - 0:  Normal Sheep conversion behaviour

        - $\geq$1:  If another player does not also have this set to a non zero value, their sheep will force convert to source player

    - Additional Information: This is how the celt bonus of sheep not converting works
    """
    OBJECT_COST_SUMMATION = 98
    """
    Total cost of all units and buildings owned by the source player
    """
    RESEARCH_COST_SUMMATION = 99
    """
    Total cost of all researches researched by the source player
    """
    RELIC_INCOME_SUMMATION = 100
    """
    Total relic gold generated by the source player
    """
    TRADE_INCOME_SUMMATION = 101
    """
    Total trade gold generated by the source player
    """
    PLAYER1_TRIBUTE = 102
    """
    Amount of resources tributed to player 1 by the source player
    """
    PLAYER2_TRIBUTE = 103
    """
    Amount of resources tributed to player 2 by the source player
    """
    PLAYER3_TRIBUTE = 104
    """
    Amount of resources tributed to player 3 by the source player
    """
    PLAYER4_TRIBUTE = 105
    """
    Amount of resources tributed to player 4 by the source player
    """
    PLAYER5_TRIBUTE = 106
    """
    Amount of resources tributed to player 5 by the source player
    """
    PLAYER6_TRIBUTE = 107
    """
    Amount of resources tributed to player 6 by the source player
    """
    PLAYER7_TRIBUTE = 108
    """
    Amount of resources tributed to player 7 by the source player
    """
    PLAYER8_TRIBUTE = 109
    """
    Amount of resources tributed to player 8 by the source player
    """
    PLAYER1_KILL_VALUE = 110
    """
    Cost of units of player 1 killed by the source player
    """
    PLAYER2_KILL_VALUE = 111
    """
    Cost of units of player 2 killed by the source player
    """
    PLAYER3_KILL_VALUE = 112
    """
    Cost of units of player 3 killed by the source player
    """
    PLAYER4_KILL_VALUE = 113
    """
    Cost of units of player 4 killed by the source player
    """
    PLAYER5_KILL_VALUE = 114
    """
    Cost of units of player 5 killed by the source player
    """
    PLAYER6_KILL_VALUE = 115
    """
    Cost of units of player 6 killed by the source player
    """
    PLAYER7_KILL_VALUE = 116
    """
    Cost of units of player 7 killed by the source player
    """
    PLAYER8_KILL_VALUE = 117
    """
    Cost of units of player 8 killed by the source player
    """
    PLAYER1_RAZINGS = 118
    """
    Number of buildings destroyed of player 1 by the source player
    """
    PLAYER2_RAZINGS = 119
    """
    Number of buildings destroyed of player 2 by the source player
    """
    PLAYER3_RAZINGS = 120
    """
    Number of buildings destroyed of player 3 by the source player
    """
    PLAYER4_RAZINGS = 121
    """
    Number of buildings destroyed of player 4 by the source player
    """
    PLAYER5_RAZINGS = 122
    """
    Number of buildings destroyed of player 5 by the source player
    """
    PLAYER6_RAZINGS = 123
    """
    Number of buildings destroyed of player 6 by the source player
    """
    PLAYER7_RAZINGS = 124
    """
    Number of buildings destroyed of player 7 by the source player
    """
    PLAYER8_RAZINGS = 125
    """
    Number of buildings destroyed of player 8 by the source player
    """
    PLAYER1_RAZING_VALUE = 126
    """
    Cost of buildings destroyed of player 1 by the source player
    """
    PLAYER2_RAZING_VALUE = 127
    """
    Cost of buildings destroyed of player 2 by the source player
    """
    PLAYER3_RAZING_VALUE = 128
    """
    Cost of buildings destroyed of player 3 by the source player
    """
    PLAYER4_RAZING_VALUE = 129
    """
    Cost of buildings destroyed of player 4 by the source player
    """
    PLAYER5_RAZING_VALUE = 130
    """
    Cost of buildings destroyed of player 5 by the source player
    """
    PLAYER6_RAZING_VALUE = 131
    """
    Cost of buildings destroyed of player 6 by the source player
    """
    PLAYER7_RAZING_VALUE = 132
    """
    Cost of buildings destroyed of player 7 by the source player
    """
    PLAYER8_RAZING_VALUE = 133
    """
    Cost of buildings destroyed of player 8 by the source player
    """
    CASTLE = 134
    """
    Number of standing castles by the source player
    """
    HIT_POINT_RAZINGS = 135
    """
    Cumulative hp of all buildings destroyed by the source player
    """
    KILLS_BY_PLAYER1 = 136
    """
    Number of units killed by player 1 of the source player
    """
    KILLS_BY_PLAYER2 = 137
    """
    Number of units killed by player 2 of the source player
    """
    KILLS_BY_PLAYER3 = 138
    """
    Number of units killed by player 3 of the source player
    """
    KILLS_BY_PLAYER4 = 139
    """
    Number of units killed by player 4 of the source player
    """
    KILLS_BY_PLAYER5 = 140
    """
    Number of units killed by player 5 of the source player
    """
    KILLS_BY_PLAYER6 = 141
    """
    Number of units killed by player 6 of the source player
    """
    KILLS_BY_PLAYER7 = 142
    """
    Number of units killed by player 7 of the source player
    """
    KILLS_BY_PLAYER8 = 143
    """
    Number of units killed by player 8 of the source player
    """
    RAZINGS_BY_PLAYER1 = 144
    """
    Number of buildings destroyed by player 1 of the source player
    """
    RAZINGS_BY_PLAYER2 = 145
    """
    Number of buildings destroyed by player 2 of the source player
    """
    RAZINGS_BY_PLAYER3 = 146
    """
    Number of buildings destroyed by player 3 of the source player
    """
    RAZINGS_BY_PLAYER4 = 147
    """
    Number of buildings destroyed by player 4 of the source player
    """
    RAZINGS_BY_PLAYER5 = 148
    """
    Number of buildings destroyed by player 5 of the source player
    """
    RAZINGS_BY_PLAYER6 = 149
    """
    Number of buildings destroyed by player 6 of the source player
    """
    RAZINGS_BY_PLAYER7 = 150
    """
    Number of buildings destroyed by player 7 of the source player
    """
    RAZINGS_BY_PLAYER8 = 151
    """
    Number of buildings destroyed by player 8 of the source player
    """
    VALUE_KILLED_BY_OTHERS = 152
    """
    Cumulative cost of units lost by the source player
    """
    VALUE_RAZED_BY_OTHERS = 153
    """
    Cumulative cost of buildings lost by the source player
    """
    KILLED_BY_OTHERS = 154
    """
    Number of units killed by other players of the source player
    """
    RAZED_BY_OTHERS = 155
    """
    Number of buildings destroyed by other players amount of the source player
    """
    TRIBUTE_FROM_PLAYER1 = 156
    """
    Tribute received from player 1 of the source player
    """
    TRIBUTE_FROM_PLAYER2 = 157
    """
    Tribute received from player 2 of the source player
    """
    TRIBUTE_FROM_PLAYER3 = 158
    """
    Tribute received from player 3 of the source player
    """
    TRIBUTE_FROM_PLAYER4 = 159
    """
    Tribute received from player 4 of the source player
    """
    TRIBUTE_FROM_PLAYER5 = 160
    """
    Tribute received from player 5 of the source player
    """
    TRIBUTE_FROM_PLAYER6 = 161
    """
    Tribute received from player 6 of the source player
    """
    TRIBUTE_FROM_PLAYER7 = 162
    """
    Tribute received from player 7 of the source player
    """
    TRIBUTE_FROM_PLAYER8 = 163
    """
    Tribute received from player 8 of the source player
    """
    VALUE_CURRENT_UNITS = 164
    """
    Cumulative cost of alive units of the source player
    """
    VALUE_CURRENT_BUILDINGS = 165
    """
    Cumulative cost of standing buildings of the source player
    """
    FOOD_TOTAL = 166
    """
    Total food collected by the source player
    """
    WOOD_TOTAL = 167
    """
    Total wood collected by the source player
    """
    STONE_TOTAL = 168
    """
    Total stone collected by the source player
    """
    GOLD_TOTAL = 169
    """
    Total gold collected by the source player
    """
    TOTAL_VALUE_OF_KILLS = 170
    """
    Cumulative cost of all units killed by the source player
    """
    TOTAL_TRIBUTE_RECEIVED = 171
    """
    Total amount of resources received in tribute by the source player
    """
    TOTAL_VALUE_OF_RAZINGS = 172
    """
    Cumulative cost of all buildings destroyed by of the source player
    """
    TOTAL_CASTLES_BUILT = 173
    """
    Number of total castles built by the source player
    """
    TOTAL_WONDERS_BUILT = 174
    """
    Number of total wonders built by the source player
    """
    TRIBUTE_SCORE = 175
    """
    Tribute score of the source player
    """
    CONVERT_MIN_ADJ = 176
    """
    Additional monk seconds needed before conversion starts for the source player

    - Additional Information: A great explanation for how this works [here](https://youtu.be/-qRUaOHpbwI?t=870 "Explanatory video by T-West") by T-West
    """
    CONVERT_MAX_ADJ = 177
    """
    Additional monk seconds needed before forced conversion for the source player

    - Additional Information: A great explanation for how this works [here](https://youtu.be/-qRUaOHpbwI?t=870 "Explanatory video by T-West") by T-West
    """
    CONVERT_RESIST_MIN_ADJ = 178
    """
    Additional monk seconds needed before conversion starts against the source player

    - Additional Information: A great explanation for how this works [here](https://youtu.be/-qRUaOHpbwI?t=830 "Explanatory video by T-West") by T-West
    """
    CONVERT_RESIST_MAX_ADJ = 179
    """
    Additional monk seconds needed before forced conversion for the source player

    - Additional Information: A great explanation for how this works [here](https://youtu.be/-qRUaOHpbwI?t=870 "Explanatory video by T-West") by T-West
    """
    CONVERT_BUILDING_MIN = 180
    """
    Building conversion min time of the source player

    - Additional Information: A great explanation for how this works [here](https://youtu.be/-qRUaOHpbwI?t=902 "Explanatory video by T-West") by T-West
    """
    CONVERT_BUILDING_MAX = 181
    """
    Building conversion max time of the source player

    - Additional Information: A great explanation for how this works [here](https://youtu.be/-qRUaOHpbwI?t=902 "Explanatory video by T-West") by T-West
    """
    CONVERT_BUILDING_CHANCE = 182
    """
    Percent chance for monks to convert buildings by the source player

    - Additional Information: A great explanation for how this works [here](https://youtu.be/-qRUaOHpbwI?t=902 "Explanatory video by T-West") by T-West
    """
    SPIES = 183
    """
    Boolean: show enemy los for the source player
    """
    VALUE_WONDERS_CASTLES = 184
    """
    Total cost of all wonders and castles of the source player
    """
    FOOD_SCORE = 185
    """
    Unknown amount of the source player
    """
    WOOD_SCORE = 186
    """
    Unknown amount of the source player
    """
    STONE_SCORE = 187
    """
    Unknown amount of the source player
    """
    GOLD_SCORE = 188
    """
    Unknown amount of the source player
    """
    WOOD_BONUS = 189
    """
    Amount of wood chopped multiplier of the source player

    - Default Values:

        - 1:  Generic

        - 1.15:  Mayans

    - Additional Information: Since this works by multiplying the amount of resources gathered by a villager, it has a side effect of increasing the gather rate as well. In the case of Mayans, This is compensated for by reducing villager work rate by 15%
    """
    FOOD_BONUS = 190
    """
    Amount of food gathered from all sources (except from sheep) multiplier of the source player

    - Default Values:

        - 1:  Generic

        - 1.15:  Mayans

    - Additional Information: Since this works by multiplying the amount of resources gathered by a villager, it has a side effect of increasing the gather rate as well. In the case of Mayans, This is compensated for by reducing villager work rate by 15%. The work rate for farmers is reduced by about 23.4%
    """
    RELIC_RATE = 191
    """
    Relic gold generation rate for the source player

    - Default Values:

        - 0.5:  by default
    """
    HERESY = 192
    """
    Boolean: converted units die for the source player
    """
    THEOCRACY = 193
    """
    Boolean: only one monk needs to regen faith after group conversion for the source player
    """
    CRENELLATIONS = 194
    """
    Boolean: researched crenellations amount of the source player

    - Additional Information: Set to 1 if crenellations is researched, does not give the effect of crenellations, just a boolean
    """
    CONSTRUCTION_RATE_MOD = 195
    """
    Builder work rate multiplier of the source player

    - Default Values:

        - 0:  Generic

        - 1.3:  Spanish

    - Additional Information: The actual work rate for builders is given by `construction_rate_mod * builder.default_work_rate`
    """
    HUN_WONDER_BONUS = 196
    """
    Additional time for relic/wonder victories in one tenth of an year by any player

    - Additional Information: Internally, relic and wonder victory countdowns are measured in one tenths of an year, the fractional part is just not shown ingame This is additive per player Set to 1000 if atheism is researched
    """
    SPIES_DISCOUNT = 197
    """
    Boolean: give discount on spies for the source player

    - Additional Information: Set to 1 if atheism is researched
    """
    UNKNOWN_RESOURCE_198 = 198
    """
    Unknown amount of source player
    """
    UNKNOWN_RESOURCE_199 = 199
    """
    Unknown amount of source player
    """
    UNKNOWN_RESOURCE_200 = 200
    """
    Unknown amount of source player
    """
    UNKNOWN_RESOURCE_201 = 201
    """
    Unknown amount of source player
    """
    UNKNOWN_RESOURCE_202 = 202
    """
    Unknown amount of source player
    """
    UNKNOWN_RESOURCE_203 = 203
    """
    Unknown amount of source player
    """
    UNKNOWN_RESOURCE_204 = 204
    """
    Unknown amount of source player
    """
    FEITORIA_FOOD_PRODUCTIVITY = 205
    """
    Feitoria/BR Trade Workshop food production multiplier of the source player

    - Default Values:

        - 1:  Generic

    - Additional Information: The amount of food obtained from owning `n` number of Feitorias (BR Trade Workshops) is given by `n * feitoria_food_productivity * 1.6 (2.25)`.
    """
    FEITORIA_WOOD_PRODUCTIVITY = 206
    """
    Feitoria/BR Trade Workshop wood production multiplier of the source player

    - Default Values:

        - 1:  Generic

    - Additional Information: The amount of wood obtained from owning `n` number of Feitorias (BR Trade Workshops) is given by `n * feitoria_wood_productivity * 1 (2.25)`.
    """
    FEITORIA_STONE_PRODUCTIVITY = 207
    """
    Feitoria/BR Trade Workshop stone production multiplier of the source player

    - Default Values:

        - 1:  Generic

    - Additional Information: The amount of stone obtained from owning `n` number of Feitorias (BR Trade Workshops) is given by `n * feitoria_stone_productivity * 0.3 (0)`.
    """
    FEITORIA_GOLD_PRODUCTIVITY = 208
    """
    Feitoria/BR Trade Workshop gold production multiplier of the source player

    - Default Values:

        - 1:  Generic

    - Additional Information: The amount of gold obtained from owning `n` number of Feitorias (BR Trade Workshops) is given by `n * feitoria_gold_productivity * 0.7 (2.25)`.
    """
    REVEAL_ENEMY_TOWN_CENTRE = 209
    """
    Boolean: reveal enemy town centre location for the source player

    - Default Values:

        - 0:  Generic

        - 5:  Vietnamese

    - Additional Information: The bonus works for all values $\geq$1, the choice of setting it to 5 for vietnamese seems arbitrary Vietnamese reveal enemy tc location bonus
    """
    REVEAL_RELICS = 210
    """
    Boolean: reveal relics on map amount of the source player

    - Default Values:

        - -1:  Generic

        - 42:  Burmese

    - Additional Information: Burmese reveal relics on map bonus (probably) Only works in RMS, Manually changing this in the editor does not seem to have any effects
    """
    ELEVATION_BONUS_HIGHER = 211
    """
    The fraction for additional bonus damage dealt from higher elevation for the source player

    - Default Values:

        - 0:  Generic

        - 0.25:  Tatars

    - Additional Information: Damage that units on higher elevation deal to units on lower elevation is multiplied by `1.25 + elevation_bonus_higher`
    """
    ELEVATION_BONUS_LOWER = 212
    """
    The fraction for additional bonus damage dealt from lower elevation for the source player

    - Default Values:

        - 0:  Generic

    - Additional Information: Damage that units on lower elevation deal to units on higher elevation is multiplied by `0.75 + elevation_bonus_lower`
    """
    RAIDING_PRODUCTIVITY = 213
    """
    Keshik gold generation rate per second*100 of the source player

    - Default Values:

        - 50:  Tatars

    - Additional Information: Note that in practice, due to attack reload time and frame delay, Keshiks don't actually produce 0.5 gold per second, but a lower value
    """
    MERCENARY_KIPCHAK_COUNT = 214
    """
    Total number of mercenary kipchak creatable by the source player

    - Additional Information: Researching Cuman Mercenaries sets this to 10. Making mercenary Kipchaks costs one unit of this resource
    """
    MERCENARY_KIPCHAK_LIMIT = 215
    """
    Number of mercenary kipchaks created/queued by the source player

    - Additional Information: Making mercenary Kipchaks gives one unit of this resource
    """
    SHEPHERD_PRODUCTIVITY = 216
    """
    Amount of food collected from sheep multiplier of the source player

    - Default Values:

        - 1:  Generic

        - 1.15:  Mayans

        - 1.57:  Tatars

    - Additional Information: Since this works by multiplying the amount of resources gathered by a villager, it has a side effect of increasing the gather rate as well. In the case of Mayans/Tatars, This is compensated for by reducing villager work rate by 15%/57%
    """
    TRIGGER_SHARED_LOS = 217
    """
    Boolean: reveal ally for the source player

    - Default Values:

        - 0:  No Allied Vision

        - 1:  Allied Vision

    - Additional Information: Note - Once enabled, Allied vison cannot be undone
    """
    UNKNOWN_RESOURCE_218 = 218
    """
    Unknown amount of the source player
    """
    UNKNOWN_RESOURCE_219 = 219
    """
    Unknown amount of the source player
    """
    UNKNOWN_RESOURCE_220 = 220
    """
    Unknown amount of the source player
    """
    FOOD_TRICKLE_FROM_MONUMENT = 221
    """
    Monument food trickle rate multiplier of source player

    - Default Values:

        - 1:  In KoTH games

    - Additional Information: The amount of resources obtained by owning a monument is `0.7925 * food_trickle_from_monument`
    """
    WOOD_TRICKLE_FROM_MONUMENT = 222
    """
    Monument wood trickle rate multiplier of source player

    - Default Values:

        - 1:  In KoTH games

    - Additional Information: The amount of resources obtained by owning a monument is `0.7925 * wood_trickle_from_monument`
    """
    STONE_TRICKLE_FROM_MONUMENT = 223
    """
    Monument stone trickle rate multiplier of source player

    - Default Values:

        - 1:  In KoTH games

    - Additional Information: The amount of resources obtained by owning a monument is `0.7925 * stone_trickle_from_monument`
    """
    GOLD_TRICKLE_FROM_MONUMENT = 224
    """
    Monument gold trickle rate multiplier of source player

    - Default Values:

        - 1:  In KoTH games

    - Additional Information: The amount of resources obtained by owning a monument is `0.7925 * gold_trickle_from_monument`
    """
    RELIC_FOOD_PRODUCTION_RATE = 225
    """
    Relic food production per minute of the source player

    - Default Values:

        - 30:  Burgundians

        - 0: Generic
    """
    VILLAGERS_KILLED_BY_GAIA = 226
    """
    Villagers lost to gaia by the source player
    """
    VILLAGERS_KILLED_BY_ANIMALS = 227
    """
    Villagers lost to wild animals by the source player
    """
    VILLAGERS_KILLED_BY_AI_PLAYER = 228
    """
    Villagers lost to ais by the source player
    """
    VILLAGERS_KILLED_BY_HUMAN_PLAYER = 229
    """
    Villagers lost to humans by the source player
    """
    FOOD_TRICKLE = 230
    """
    Food given per minute to the source player
    """
    WOOD_TRICKLE = 231
    """
    Wood given per minute to the source player
    """
    STONE_TRICKLE = 232
    """
    Stone given per minute to the source player
    """
    GOLD_TRICKLE = 233
    """
    Gold given per minute to the source player
    """
    SPAWN_LIMIT = 234
    """
    Limit of the number of spawning buildings that spawn units from spawn command in a technology of the source player
    """
    FLEMISH_MILITIA_POPULATION = 235
    """
    Number of flemish militia of the source player
    """
    FARMING_GOLD_TRICKLE = 236
    """
    Farming gold generation rate per second*100 of the source player

    - Additional Information: used by the Burgundian Vineyards technology. Set to 2 when researched
    """
    FOLWARK_COLLECTION_AMOUNT = 237
    """
    This is the amount of food collected from farms built around a Folwark by the source player

    - Default Values:

        - 0:  Generic

        - 17.5:  Poles

        - 25:  Poles with Horse Collar

        - 37.5:  Poles with Heavy Plow

        - 55:  Poles with Crop Rotation

    - Additional Information: Poles set this to 17.5. Horse Collar adds 7.5 to it, Heavy Plow adds 12.5 to it and Crop Rotation adds 17.5 to it
    """
    FOLWARK_ATTRIBUTE_TYPE = 238
    """
    This is the ID of the resource that is given to the player from a farm constructed around a Folwark to the source player

    - Default Values:

        - 0:  Poles

        - -1:  Generic
    """
    FOLWARK_BUILDING_TYPE = 239
    """
    This is the ID of the building that the Folwark needs to upgrade from for the farm collection ability to work for the source player

    - Default Values:

        - 68 (Mill):  Poles

        - -1:  Generic
    """
    UNITS_CONVERTED = 240
    """
    The amount of units lost to enemy conversions by the source player
    """
    STONE_MINING_GOLD_PRODUCTIVITY = 241
    """
    Stone mining gold generation rate per second*100 of the source player

    - Default Values:

        - 0:  Generic

        - 18:  Poles

        - 20.7:  Poles with Stone Mining

        - 23.805:  Poles with Stone Shaft Mining
    """
    NUMBER_OF_FREE_TRANSPORTS = 242
    """
    This is the number of free transport ships trainable by having a Sicilian ally to the source player

    - Default Values:

    	- 1:  Generic

    - Additional Information: One unit of this resource is required to be able to train the free transport ship granted by having a Sicilian ally. This seemingly useless cost is necessary to prevent shift queuing multiple of the free transport ship unit. Note that as soon as at least one free transport ship is created, the unit disables itself, no matter how much of this resource is left. It is possible to change this by disabling technology 229, but then keep in mind that regular transport ships won't be trainable once this resource runs out
    """


class ObjectType(_DataSetIntEnums):
    """
    This enum class provides the integer values used to reference the object types in the game. Used in a lot of effects
    and conditions, like 'Kill Object', 'Objects in Area'.

    **Examples**

    >>> ObjectType.OTHER
    <ObjectType.OTHER: 1>
    """
    OTHER = 1
    BUILDING = 2
    CIVILIAN = 3
    MILITARY = 4


class ObjectClass(_DataSetIntEnums):
    """
    This enum class provides the integer values used to reference the object class in the game. Used in a lot of effects
    and conditions, like 'Kill Object', 'Objects in Area' under the name 'Object Group'.

    **Examples**

    >>> ObjectClass.CIVILIAN
    <ObjectClass.CIVILIAN: 4>
    """
    ARCHER = 0
    ARTIFACT = 1
    TRADE_BOAT = 2
    BUILDING = 3
    CIVILIAN = 4
    OCEAN_FISH = 5
    INFANTRY = 6
    BERRY_BUSH = 7
    STONE_MINE = 8
    PREY_ANIMAL = 9
    PREDATOR_ANIMAL = 10
    MISCELLANEOUS = 11
    CAVALRY = 12
    SIEGE_WEAPON = 13
    TERRAIN = 14
    TREE = 15
    TREE_STUMP = 16
    HEALER = 17
    MONK = 18
    TRADE_CART = 19
    TRANSPORT_BOAT = 20
    FISHING_BOAT = 21
    WARSHIP = 22
    CONQUISTADOR = 23
    WAR_ELEPHANT = 24
    HERO = 25
    ELEPHANT_ARCHER = 26
    WALL = 27
    PHALANX = 28
    DOMESTIC_ANIMAL = 29
    FLAG = 30
    DEEP_SEA_FISH = 31
    GOLD_MINE = 32
    SHORE_FISH = 33
    CLIFF = 34
    PETARD = 35
    CAVALRY_ARCHER = 36
    DOPPELGANGER = 37
    BIRD = 38
    GATE = 39
    SALVAGE_PILE = 40
    RESOURCE_PILE = 41
    RELIC = 42
    MONKWITH_RELIC = 43
    HAND_CANNONEER = 44
    TWO_HANDED_SWORDSMAN = 45
    PIKEMAN = 46
    SCOUT = 47
    ORE_MINE = 48
    FARM = 49
    SPEARMAN = 50
    PACKED_UNIT = 51
    TOWER = 52
    BOARDING_BOAT = 53
    UNPACKED_SIEGE_UNIT = 54
    BALLISTA = 55
    RAIDER = 56
    CAVALRY_RAIDER = 57
    LIVESTOCK = 58
    KING = 59
    MISC_BUILDING = 60
    CONTROLLED_ANIMAL = 61


class TerrainRestrictions(_DataSetIntEnums):
    """
    This enum class provides the integer values used to reference the terrain restriction IDs in the game. Used in the
    'Modify Attribute' effects

    **Examples**

    >>> TerrainRestrictions.LAND_AND_SHALLOWS
    <TerrainRestrictions.LAND_AND_SHALLOWS: 1>
    """
    ALL = 0
    """Used by terrain eyecandy and sundries"""
    LAND_AND_SHALLOWS = 1
    """Used by part of animals"""
    BEACH = 2
    WATER_SMALL_TRAIL = 3
    """Used by most ships and sea gate"""
    LAND = 4
    """Used by most land buildings"""
    NOTHING = 5
    WATER_NO_TRAIL = 6
    """Used by docks"""
    ALL_EXCEPT_WATER = 7
    """Used by land troops"""
    LAND_EXCEPT_FARM = 8
    """Used by land resources"""
    NOTHING_2 = 9
    LAND_AND_BEACH = 10
    """Used by land walls and gates"""
    LAND_EXCEPT_FARM_2 = 11
    """Used by trees and mountains"""
    ALL_EXCEPT_WATER_BRIDGE_CANNON = 12
    WATER_MEDIUM_TRAIL = 13
    """Used by big fish and fishing ship"""
    ALL_EXCEPT_WATER_BRIDGE_ARROW = 14
    WATER_LARGE_TRAIL = 15
    """Only used by transport ship"""
    GRASS_AND_BEACH = 16
    WATER_AND_BRIDGE_EXCEPT_BEACH = 17
    ALL_EXCEPT_WATER_BRIDGE_SPEAR = 18
    ONLY_WATER_AND_ICE = 19
    """Used by fish"""
    ALL_EXCEPT_WATER_WHEEL = 20
    """Used by units with wheels, such as Rams and Scorpions"""
    SHALLOW_WATER = 21
    ALL_DART = 22
    ALL_ARROW_FIRE = 23
    """Only used by Arrows with fire (After chemistry)"""
    ALL_CANNON_FIRE = 24
    """Only used by Cannon balls (After chemistry)"""
    ALL_SPEAR_FIRE = 25
    """Only used by Spears with fire (After chemistry)"""
    ALL_DART_FIRE = 26
    """Only used by Darts with fire (After chemistry)"""
    ALL_LASER = 27
    """Only used by Projectile Laser with id 1595"""
    ALL_EXCEPT_WATER_CAVALRY = 28
    """Such as Cavalry Archer, Cavalry, Conquistador, Missionary and Flaming Camel"""
    ALL_EXCEPT_WATER_PACKET_TREBUCHET = 29
    """All types of Trebuchet(Packed)"""
    WATER_SMALLEST_TRAIL = 30
    """Used by medium ships, such as Trade Cog, Fire Galley and Longboat"""


class HeroStatusFlag(_DataSetIntFlags):
    """
    This enum class provides the integer values for the different hero status flags that can be used in the 'Modify
    Attribute' effect with the 'Hero Status' attribute.

    **Methods**

    - ``HeroStatusFlag.combine()``
    - ``HeroStatusFlag.split_flags()``


    **Examples**

    >>> HeroStatusFlag.HERO_REGENERATION
    <HeroStatusFlag.HERO_REGENERATION: 4>
    >>> HeroStatusFlag.HERO_REGENERATION | HeroStatusFlag.HERO_GLOW
    <HeroStatusFlag.HERO_GLOW|HERO_REGENERATION: 68>
    >>> HeroStatusFlag.combine(hero_regeneration=True, hero_glow=True)
    <HeroStatusFlag.HERO_GLOW|HERO_REGENERATION: 68>
    """

    @staticmethod
    def combine(
            full_hero_status: bool = False,
            cannot_be_converted: bool = False,
            hero_regeneration: bool = False,
            defensive_stance_by_default: bool = False,
            protected_formation: bool = False,
            delete_confirmation: bool = False,
            hero_glow: bool = False,
            invert_all_flags: bool = False
    ) -> HeroStatusFlag:
        """
        This method combines the given hero status flags into an integer value

        Args:
            full_hero_status: Enabling this for a unit grants all the flags mentioned below except invert_all_flags
            cannot_be_converted: Enabling this for a unit makes it un-convertable
            hero_regeneration: Enabling this for a unit grants 0.5 HP/s heal rate to the unit
            defensive_stance_by_default: Enabling this for a unit makes it be on defensive stance by default
            protected_formation: Enabling this for a unit makes it be in protected formation by default
            delete_confirmation: Enabling this for a unit will bring up a delete confirmation for the unit when trying
            to delete it IF the player has them enabled
            hero_glow: Enabling this for a unit grants it the golden hero glow effect
            invert_all_flags: Enabling this for a unit will invert all the above flags except full_hero_status

        Returns:
            An integer combining all the different hero status flags into one value
        """
        total = 1 if full_hero_status else 0
        total += 2 if cannot_be_converted else 0
        total += 4 if hero_regeneration else 0
        total += 8 if defensive_stance_by_default else 0
        total += 16 if protected_formation else 0
        total += 32 if delete_confirmation else 0
        total += 64 if hero_glow else 0
        total += 128 if invert_all_flags else 0
        return HeroStatusFlag(total)

    @staticmethod
    def split_flags(value: int) -> dict[HeroStatusFlag, bool]:
        """
        Split the Hero Status flags into boolean variables related to their effects

        Args:
            value: An integer value representing all the hero status flags set

        Returns:
            A dict with all the flags values as keys and a bool as their value
        """
        flags = {}
        for flag in HeroStatusFlag:
            flags[flag] = bool(flag & value)

        return flags

    FULL_HERO_STATUS = 1
    CANNOT_BE_CONVERTED = 2
    HERO_REGENERATION = 4
    DEFENSIVE_STANCE_BY_DEFAULT = 8
    PROTECTED_FORMATION = 16
    DELETE_CONFIRMATION = 32
    HERO_GLOW = 64
    INVERT_FLAGS = 128


class BlastLevel(_DataSetIntEnums):
    """
    This enum class provides the integer values used to reference the blast level values used in the game. Used in the
    'Modify Attribute' effect with the 'Blast Attack/Defense Level' attributes

    **Examples**

    >>> BlastLevel.TREES
    <BlastLevel.TREES: 1>
    """
    RESOURCES = 0
    """
    Damage Resources
    """
    TREES = 1
    """
    Damage Trees
    """
    NEARBY_UNITS = 2
    """
    Damage surrounding units
    """
    TARGET_ONLY = 3
    """
    Damage targetted unit only
    """
    FIXED_FIVE = 4
    """
    Deal a fixed, 5 HP of damage to surrounding units
    """
    FIFTY_PERCENT = 8
    """
    Deal 50% damage to surrounding units
    """
    TWNETY_FIVE_PERCENT = 16
    """
    Deal 25% damage to surrounding units
    """
    THIRTY_THREE_PERCENT = 32
    """
    Deal 33% damage to surrounding units
    """
    DISTANCE_ATTENUATION = 64
    """
    Only works for infantry units, reduces blast damage done based on distance
    """


class SmartProjectile(_DataSetIntEnums):
    """
    This enum class provides the integer values used to reference the smart projectile flag values used in the game. Used in the
    'Modify Attribute' effect with the 'Enable Smart Projectile' attribute

    **Examples**

    >>> SmartProjectile.ENABLED
    <SmartProjectile.ENABLED: 1>
    """
    ENABLED = 1
    FULL_DAMAGE_ON_MISSED_HIT = 2


class DamageClass(_DataSetIntEnums):
    """
    This enum class provides the integer values that represent the damage classes in the game. Used in the 'Chnage
    Object Attack/Armour' and 'Modify Attribute' with the 'Attack/Armour' attibutes

    **Examples**

    >>> DamageClass.INFANTRY
    <DamageClass.INFANTRY: 1>
    """
    WONDER = 0
    """Since HD. Only wonders has this armour class. However there is no unit that has this attack class."""
    INFANTRY = 1
    TURTLE_SHIPS = 2
    BASE_PIERCE = 3
    BASE_MELEE = 4
    WAR_ELEPHANTS = 5
    UNUSED_ID6 = 6
    UNUSED_ID7 = 7
    CAVALRY = 8
    UNUSED_ID9 = 9
    UNUSED_ID10 = 10
    ALL_BUILDINGS_EXCEPT_PORT = 11
    """Port is the building with id 446"""
    UNUSED_ID12 = 12
    STONE_DEFENSE = 13
    PREDATOR_ANIMALS_FE = 14
    """Wolf, Bear, Jaguar, Tiger, etc. have this armour class"""
    ARCHERS = 15
    BATTLE_SHIPS_AND_SABOTEUR = 16
    """Camels also had this armour class before AK"""
    RAMS = 17
    TREES = 18
    UNIQUE_UNITS = 19
    SIEGE_WEAPONS = 20
    STANDARD_BUILDINGS = 21
    WALLS_AND_GATES = 22
    GUNPOWDER_UNITS = 23
    BOARS = 24
    MONKS = 25
    CASTLE = 26
    SPEARMEN = 27
    CAVALRY_ARCHERS = 28
    EAGLE_WARRIORS = 29
    CAMELS = 30
    """Camels use this armour class since and after AK"""
    LEITIS = 31
    CONDOTTIERO = 32
    ORGAN_GUN_BULLET = 33
    """Only Projectile Gunpowder (Secondary) with id 1119 has this damage class. No unit has this armour class."""
    FISHING_SHIP = 34
    MAMELUKES = 35
    HEROES_AND_KING = 36
    UNUSED_ID37 = 37
    UNUSED_ID38 = 38
    UNUSED_ID39 = 39
    UNUSED_ID40 = 40
    UNUSED_ID41 = 41
    UNUSED_ID42 = 42
    UNUSED_ID43 = 43
    UNUSED_ID44 = 44
    UNUSED_ID45 = 45
    UNUSED_ID46 = 46
    UNUSED_ID47 = 47
    UNUSED_ID48 = 48
    UNUSED_ID49 = 49


class Hotkey(_DataSetIntEnums):
    """
    This enum class provides the integer values used to reference the blast level values used in the game. Used in the
    'Modify Attribute' effect with the 'HotKey ID' attribute

    Many hotkeys are missing from this file (Like arrow up). The reason for this is explained in the UGC guide:
    https://divy1211.github.io/AoE2DE_UGC_Guide/general/hotkeys/hotkeys/

    **Examples**

    >>> Hotkey.SPACE
    <Hotkey.SPACE: 10101>
    """
    SPACE = 10101
    PAGE_UP = 15000
    LEFT_ARROW = 2312
    RIGHT_ARROW = 19707
    DOWN_ARROW = 19731
    INSERT = 9025
    DELETE = 19602
    ZERO = 99
    ONE = 98
    TWO = 10360
    THREE = 9786
    FOUR = 10362
    FIVE = 9785
    SIX = 213
    SEVEN = 8828
    EIGHT = 9448
    NINE = 9783
    A = 1001
    B = 1005
    C = 1201
    D = 1151
    E = 1007
    F = 4137
    G = 2012
    H = 2407
    I = 1212
    J = 1222
    K = 4141
    L = 1101
    M = 1006
    N = 3001
    O = 214
    P = 1210
    Q = 4169
    R = 1200
    S = 1102
    T = 2016
    U = 1205
    V = 1150
    W = 1008
    X = 1002
    Y = 2008
    Z = 4174
    APPLICATION = 19704
    NUM_ZERO = 19721
    NUM_ONE = 4563
    NUM_FOUR = 19499
    NUM_FIVE = 5558
    NUM_SEVEN = 10069
    NUM_EIGHT = 1011
    NUM_DELETE = 20123
    F3 = 9798
    F4 = 22019
    F7 = 9840
    F8 = 1152
    F15 = 10661


class ColorMood(_DataSetIntEnums):
    """
    This enum class provides the integer values used to reference the color mood values used in the game. Used in the
    'Change Color Mood' effect

    **Examples**

    >>> ColorMood.AUTUMN
    <ColorMood.AUTUMN: 1>
    """
    DEFAULT = 0
    AUTUMN = 1
    WINTER = 2
    JUNGLE = 3
    DESERT = 4
    NIGHT = 5


class ObjectState(_DataSetIntEnums):
    """
    This enum class provides the integer values used to reference the object state values used in the game. Used in the
    'Object in Area' condition

    **Examples**

    >>> BlastLevel.TREES
    <BlastLevel.TREES: 1>
    """
    FOUNDATION = 0
    ALMOST_ALIVE = 1
    ALIVE = 2
    RESOURCE = 3
    DYING = 4
    DEAD = 5
    UNDEAD = 6
    REMOVE = 7


class Age(_DataSetIntEnums):
    """
    This enum class provides the integer values used to reference the different ages in the game. These values are
    used by the 'Current Age' player resource

    **Examples**

    >>> Age.IMPERIAL_AGE
    <Age.IMPERIAL_AGE: 3>
    """

    DARK_AGE = 0
    FEUDAL_AGE = 1
    CASTLE_AGE = 2
    IMPERIAL_AGE = 3


class ActionType(_DataSetIntEnums):
    """
    This enum class provides the integer values used to reference the different action types in the game.
    These values are used in the Task Object effect

    **Examples**

    >>> ActionType.DROP_RELIC
    <ActionType.DROP_RELIC: 14>
    """
    DEFAULT = 0
    MOVE = 1
    PATROL = 2
    GUARD = 3
    FOLLOW = 4
    STOP = 5
    ATTACK_GROUND = 6
    GARRISON = 7
    KILL = 8
    UNLOAD = 9
    GATHER_POINT = 10
    LOCK_UNLOCK = 11
    WORK = 12
    UNGARRISON = 13
    DROP_RELIC = 14
    PACK = 15
    UNPACK = 16
    ATTACK_MOVE = 17


class SecondaryGameMode(_DataSetIntFlags):
    """
    This enum class provides the integer values for the different secondary game modes.

    **Examples**

    >>> SecondaryGameMode.EMPIRE_WARS
    <SecondaryGameMode.EMPIRE_WARS: 1>
    >>> SecondaryGameMode.SUDDEN_DEATH | SecondaryGameMode.REGICIDE
    <HeroStatusFlag.SUDDEN_DEATH|REGICIDE: 6>
    """

    @staticmethod
    def combine(
            empire_wars: bool = False,
            sudden_death: bool = False,
            regicide: bool = False,
            king_of_the_hill: bool = False,
    ) -> SecondaryGameMode:
        """
        This method combines the given hero status flags into an integer value

        Args:
            empire_wars (bool): If empire wars should be enabled
            sudden_death (bool): If sudden death should be enabled
            regicide (bool): If regicide should be enabled
            king_of_the_hill (bool): If king of the hill should be enabled

        Returns:
            An integer combining all the different hero status flags into one value
        """
        total = 1 if empire_wars else 0
        total += 2 if sudden_death else 0
        total += 4 if regicide else 0
        total += 8 if king_of_the_hill else 0
        return SecondaryGameMode(total)

    EMPIRE_WARS = 1
    SUDDEN_DEATH = 2
    REGICIDE = 4
    KING_OF_THE_HILL = 8



from enum import IntEnum


class EffectId(IntEnum):
    """
    This enum class provides the integer values used to reference the effects in the game. Used in every effect to
    indicate which type of effect it is

    **Examples**

    >>> EffectId.RESEARCH_TECHNOLOGY
    >>> 2
    """
    NONE = 0
    """Attributes for the **none** effect are: \n
    ... none... Just like Conditions... People these days... """
    CHANGE_DIPLOMACY = 1
    """Attributes for the **change_diplomacy** effect are: \n
    - diplomacy
    - source_player
    - target_player"""
    RESEARCH_TECHNOLOGY = 2
    """Attributes for the **research_technology** effect are: \n
    - source_player
    - technology
    - force_research_technology"""
    SEND_CHAT = 3
    """Attributes for the **send_chat** effect are: \n
    - source_player
    - string_id
    - message
    - sound_name"""
    PLAY_SOUND = 4
    """Attributes for the **play_sound** effect are: \n
    - source_player
    - location_x
    - location_y
    - location_object_reference
    - sound_name"""
    TRIBUTE = 5
    """Attributes for the **tribute** effect are: \n
    - quantity
    - tribute_list
    - source_player
    - target_player"""
    UNLOCK_GATE = 6
    """Attributes for the **unlock_gate** effect are: \n
    - selected_object_ids"""
    LOCK_GATE = 7
    """Attributes for the **lock_gate** effect are: \n
    - selected_object_ids"""
    ACTIVATE_TRIGGER = 8
    """Attributes for the **activate_trigger** effect are: \n
    - trigger_id"""
    DEACTIVATE_TRIGGER = 9
    """Attributes for the **deactivate_trigger** effect are: \n
    - trigger_id"""
    AI_SCRIPT_GOAL = 10
    """Attributes for the **deactivate_trigger** effect are: \n
    - ai_script_goal"""
    CREATE_OBJECT = 11
    """Attributes for the **create_object** effect are: \n
    - object_list_unit_id
    - source_player
    - location_x
    - location_y
    - facet"""
    TASK_OBJECT = 12
    """Attributes for the **task_object** effect are: \n
    - object_list_unit_id
    - source_player
    - location_x
    - location_y
    - location_object_reference
    - area_x1
    - area_y1
    - area_x2
    - area_y2
    - object_group
    - object_type
    - action_type
    - selected_object_ids"""
    DECLARE_VICTORY = 13
    """Attributes for the **declare_victory** effect are: \n
    - source_player
    - enabled"""
    KILL_OBJECT = 14
    """Attributes for the **kill_object** effect are: \n
    - object_list_unit_id
    - source_player
    - area_x1
    - area_y1
    - area_x2
    - area_y2
    - object_group
    - object_type
    - selected_object_ids"""
    REMOVE_OBJECT = 15
    """Attributes for the **remove_object** effect are: \n
    - object_list_unit_id
    - source_player
    - area_x1
    - area_y1
    - area_x2
    - area_y2
    - object_group
    - object_type
    - object_state
    - selected_object_ids"""
    CHANGE_VIEW = 16
    """Attributes for the **change_view** effect are: \n
    - source_player
    - location_x
    - location_y
    - scroll"""
    UNLOAD = 17
    """Attributes for the **unload** effect are: \n
    - object_list_unit_id
    - source_player
    - location_x
    - location_y
    - location_object_reference
    - area_x1
    - area_y1
    - area_x2
    - area_y2
    - object_group
    - object_type
    - selected_object_ids"""
    CHANGE_OWNERSHIP = 18
    """Attributes for the **change_ownership** effect are: \n
    - object_list_unit_id
    - source_player
    - target_player
    - area_x1
    - area_y1
    - area_x2
    - area_y2
    - object_group
    - object_type
    - flash_object
    - selected_object_ids"""
    PATROL = 19
    """Attributes for the **patrol** effect are: \n
    - object_list_unit_id
    - source_player
    - location_x
    - location_y
    - location_object_reference
    - area_x1
    - area_y1
    - area_x2
    - area_y2
    - object_group
    - object_type
    - selected_object_ids"""
    DISPLAY_INSTRUCTIONS = 20
    """Attributes for the **display_instructions** effect are: \n
    - object_list_unit_id
    - source_player
    - string_id
    - display_time
    - instruction_panel_position
    - play_sound
    - message
    - sound_name"""
    CLEAR_INSTRUCTIONS = 21
    """Attributes for the **clear_instructions** effect are: \n
    - instruction_panel_position"""
    FREEZE_OBJECT = 22
    """Attributes for the **freeze_object** effect are: \n
    - object_list_unit_id
    - source_player
    - area_x1
    - area_y1
    - area_x2
    - area_y2
    - object_group
    - object_type
    - selected_object_ids"""
    USE_ADVANCED_BUTTONS = 23
    """Attributes for the **use_advanced_buttons** effect are: \n
    None. \n
    Please don't use this effect. Please."""
    DAMAGE_OBJECT = 24
    """Attributes for the **damage_object** effect are: \n
    - quantity
    - object_list_unit_id
    - source_player
    - area_x1
    - area_y1
    - area_x2
    - area_y2
    - object_group
    - object_type
    - selected_object_ids"""
    PLACE_FOUNDATION = 25
    """Attributes for the **place_foundation** effect are: \n
    - object_list_unit_id
    - source_player
    - location_x
    - location_y"""
    CHANGE_OBJECT_NAME = 26
    """Attributes for the **change_object_name** effect are: \n
    - object_list_unit_id
    - source_player
    - string_id
    - area_x1
    - area_y1
    - area_x2
    - area_y2
    - message
    - selected_object_ids"""
    CHANGE_OBJECT_HP = 27
    """Attributes for the **change_object_hp** effect are: \n
    - quantity
    - object_list_unit_id
    - source_player
    - area_x1
    - area_y1
    - area_x2
    - area_y2
    - object_group
    - object_type
    - operation
    - selected_object_ids"""
    CHANGE_OBJECT_ATTACK = 28
    """Attributes for the **change_object_attack** effect are: \n
    - armour_attack_quantity
    - armour_attack_class
    - object_list_unit_id
    - source_player
    - area_x1
    - area_y1
    - area_x2
    - area_y2
    - object_group
    - object_type
    - operation
    - selected_object_ids"""
    STOP_OBJECT = 29
    """Attributes for the **stop_object** effect are: \n
    - object_list_unit_id
    - source_player
    - area_x1
    - area_y1
    - area_x2
    - area_y2
    - object_group
    - object_type
    - selected_object_ids"""
    ATTACK_MOVE = 30
    """Attributes for the **attack_move** effect are: \n
    - object_list_unit_id
    - source_player
    - location_x
    - location_y
    - location_object_reference
    - area_x1
    - area_y1
    - area_x2
    - area_y2
    - object_group
    - object_type
    - selected_object_ids"""
    CHANGE_OBJECT_ARMOR = 31
    """Attributes for the **change_object_armor** effect are: \n
    - armour_attack_quantity
    - armour_attack_class
    - object_list_unit_id
    - source_player
    - area_x1
    - area_y1
    - area_x2
    - area_y2
    - object_group
    - object_type
    - operation
    - selected_object_ids"""
    CHANGE_OBJECT_RANGE = 32
    """Attributes for the **change_object_range** effect are: \n
    - quantity
    - object_list_unit_id
    - source_player
    - area_x1
    - area_y1
    - area_x2
    - area_y2
    - object_group
    - object_type
    - operation
    - selected_object_ids"""
    CHANGE_OBJECT_SPEED = 33
    """Attributes for the **change_object_speed** effect are: \n
    - quantity
    - object_list_unit_id
    - source_player
    - area_x1
    - area_y1
    - area_x2
    - area_y2
    - object_group
    - object_type
    - selected_object_ids"""
    HEAL_OBJECT = 34
    """Attributes for the **heal_object** effect are: \n
    - quantity
    - object_list_unit_id
    - source_player
    - area_x1
    - area_y1
    - area_x2
    - area_y2
    - object_group
    - object_type
    - selected_object_ids"""
    TELEPORT_OBJECT = 35
    """Attributes for the **teleport_object** effect are: \n
    - object_list_unit_id
    - source_player
    - location_x
    - location_y
    - area_x1
    - area_y1
    - area_x2
    - area_y2
    - object_group
    - object_type
    - selected_object_ids"""
    CHANGE_OBJECT_STANCE = 36
    """Attributes for the **change_object_stance** effect are: \n
    - object_list_unit_id
    - source_player
    - area_x1
    - area_y1
    - area_x2
    - area_y2
    - object_group
    - object_type
    - attack_stance
    - selected_object_ids"""
    DISPLAY_TIMER = 37
    """Attributes for the **display_timer** effect are: \n
    - string_id
    - display_time
    - time_unit
    - timer
    - reset_timer
    - message"""
    ENABLE_DISABLE_OBJECT = 38
    """Attributes for the **enable_disable_object** effect are: \n
    - object_list_unit_id
    - source_player
    - enabled"""
    ENABLE_DISABLE_TECHNOLOGY = 39
    """Attributes for the **enable_disable_technology** effect are: \n
    - source_player
    - technology
    - enabled"""
    CHANGE_OBJECT_COST = 40
    """Attributes for the **change_object_cost** effect are: \n
    - object_list_unit_id
    - source_player
    - food
    - wood
    - stone
    - gold"""
    SET_PLAYER_VISIBILITY = 41
    """Attributes for the **set_player_visibility** effect are: \n
    - source_player
    - target_player
    - visibility_state"""
    CHANGE_OBJECT_ICON = 42
    """Attributes for the **change_object_icon** effect are: \n
    - object_list_unit_id
    - source_player
    - area_x1
    - area_y1
    - area_x2
    - area_y2
    - object_group
    - object_type
    - object_list_unit_id_2
    - selected_object_ids"""
    REPLACE_OBJECT = 43
    """Attributes for the **replace_object** effect are: \n
    - object_list_unit_id
    - source_player
    - target_player
    - area_x1
    - area_y1
    - area_x2
    - area_y2
    - object_group
    - object_type
    - object_list_unit_id_2
    - selected_object_ids"""
    CHANGE_OBJECT_DESCRIPTION = 44
    """Attributes for the **change_object_description** effect are: \n
    - object_list_unit_id
    - source_player
    - string_id
    - message"""
    CHANGE_PLAYER_NAME = 45
    """Attributes for the **change_player_name** effect are: \n
    - source_player
    - string_id
    - message"""
    CHANGE_TRAIN_LOCATION = 46
    """Attributes for the **change_train_location** effect are: \n
    - object_list_unit_id
    - source_player
    - object_list_unit_id_2
    - button_location"""
    CHANGE_RESEARCH_LOCATION = 47
    """Attributes for the **change_research_location** effect are: \n
    - source_player
    - technology
    - object_list_unit_id_2
    - button_location"""
    CHANGE_CIVILIZATION_NAME = 48
    """Attributes for the **change_civilization_name** effect are: \n
    - source_player
    - string_id
    - message"""
    CREATE_GARRISONED_OBJECT = 49
    """Attributes for the **create_garrisoned_object** effect are: \n
    - object_list_unit_id
    - source_player
    - area_x1
    - area_y1
    - area_x2
    - area_y2
    - object_list_unit_id_2
    - selected_object_ids"""
    ACKNOWLEDGE_AI_SIGNAL = 50
    """Attributes for the **acknowledge_ai_signal** effect are: \n
    - ai_signal_value"""
    MODIFY_ATTRIBUTE = 51
    """Attributes for the **modify_attribute** effect are: \n
    - quantity
    - object_list_unit_id
    - source_player
    - operation
    - object_attributes
    - message"""
    MODIFY_RESOURCE = 52
    """Attributes for the **modify_resource** effect are: \n
    - quantity
    - tribute_list
    - source_player
    - operation"""
    MODIFY_RESOURCE_BY_VARIABLE = 53
    """Attributes for the **modify_resource_by_variable** effect are: \n
    - tribute_list
    - source_player
    - operation
    - variable"""
    SET_BUILDING_GATHER_POINT = 54
    """Attributes for the **set_building_gather_point** effect are: \n
    - object_list_unit_id
    - source_player
    - location_x
    - location_y
    - area_x1
    - area_y1
    - area_x2
    - area_y2
    - selected_object_ids"""
    SCRIPT_CALL = 55
    """Attributes for the **script_call** effect are: \n
    - string_id
    - message"""
    CHANGE_VARIABLE = 56
    """Attributes for the **change_variable** effect are: \n
    - quantity
    - operation
    - variable
    - message"""
    CLEAR_TIMER = 57
    """Attributes for the **clear_timer** effect are: \n
    - timer"""
    CHANGE_OBJECT_PLAYER_COLOR = 58
    """Attributes for the **change_object_player_color** effect are: \n
    - object_list_unit_id
    - source_player
    - area_x1
    - area_y1
    - area_x2
    - area_y2
    - player_color
    - selected_object_ids"""
    CHANGE_OBJECT_CIVILIZATION_NAME = 59
    """Attributes for the **change_object_civilization_name** effect are: \n
    - string_id
    - area_x1
    - area_y1
    - area_x2
    - area_y2
    - selected_object_ids"""
    CHANGE_OBJECT_PLAYER_NAME = 60
    """Attributes for the **change_object_player_name** effect are: \n
    - object_list_unit_id
    - source_player
    - string_id
    - area_x1
    - area_y1
    - area_x2
    - area_y2
    - selected_object_ids"""
    DISABLE_UNIT_TARGETING = 61
    """Attributes for the **disable_unit_targeting** effect are: \n
    - object_list_unit_id
    - source_player
    - area_x1
    - area_y1
    - area_x2
    - area_y2
    - selected_object_ids"""
    ENABLE_UNIT_TARGETING = 62
    """Attributes for the **enable_unit_targeting** effect are: \n
    - object_list_unit_id
    - source_player
    - area_x1
    - area_y1
    - area_x2
    - area_y2
    - selected_object_ids"""
    CHANGE_TECHNOLOGY_COST = 63
    """Attributes for the **change_technology_cost** effect are: \n
    - source_player
    - technology
    - food
    - wood
    - stone
    - gold"""
    CHANGE_TECHNOLOGY_RESEARCH_TIME = 64
    """Attributes for the **change_technology_research_time** effect are: \n
    - quantity
    - source_player
    - technology"""
    CHANGE_TECHNOLOGY_NAME = 65
    """Attributes for the **change_technology_name** effect are: \n
    - source_player
    - technology
    - string_id
    - message"""
    CHANGE_TECHNOLOGY_DESCRIPTION = 66
    """Attributes for the **change_technology_description** effect are: \n
    - source_player
    - technology
    - string_id
    - message"""
    ENABLE_TECHNOLOGY_STACKING = 67
    """Attributes for the **enable_technology_stacking** effect are: \n
    - source_player
    - technology"""
    DISABLE_TECHNOLOGY_STACKING = 68
    """Attributes for the **disable_technology_stacking** effect are: \n
    - source_player
    - technology"""
    ACKNOWLEDGE_MULTIPLAYER_AI_SIGNAL = 69
    """Attributes for the **acknowledge_multiplayer_ai_signal** effect are: \n
    - ai_signal_value
    """
    DISABLE_OBJECT_SELECTION = 70
    """Attributes for the **disable_object_selection** effect are: \n
    - object_list_unit_id
    - source_player
    - area_x1
    - area_y1
    - area_x2
    - area_y2
    - selected_object_ids"""
    ENABLE_OBJECT_SELECTION = 71
    """Attributes for the **enable_object_selection** effect are: \n
    - object_list_unit_id
    - source_player
    - area_x1
    - area_y1
    - area_x2
    - area_y2
    - selected_object_ids"""
    CHANGE_COLOR_MOOD = 72
    """Attributes for the **change_color_mood** effect are: \n
    - quantity
    - color_mood
    
    **Version notice**: \n
    This condition is supported since: 1.42"""
    # Possibly used for marking when achievements should be achieved. Doesn't seem to work in scenario's. F
    # UNKNOWN_0 = 58  # Was 58... Not anymore (?)


empty_attributes = {
    "effect_type": -1,
    "ai_script_goal": -1,
    "armour_attack_quantity": -1,
    "armour_attack_class": -1,
    "quantity": -1,
    "tribute_list": -1,
    "diplomacy": -1,
    "object_list_unit_id": -1,
    "source_player": -1,
    "target_player": -1,
    "technology": -1,
    "string_id": -1,
    "display_time": -1,
    "trigger_id": -1,
    "location_x": -1,
    "location_y": -1,
    "location_object_reference": -1,
    "area_x1": -1,
    "area_y1": -1,
    "area_x2": -1,
    "area_y2": -1,
    "object_group": -1,
    "object_type": -1,
    "instruction_panel_position": -1,
    "attack_stance": -1,
    "time_unit": -1,
    "enabled": -1,
    "food": -1,
    "wood": -1,
    "stone": -1,
    "gold": -1,
    "item_id": -1,
    "flash_object": -1,
    "force_research_technology": -1,
    "visibility_state": -1,
    "scroll": -1,
    "operation": -1,
    "object_list_unit_id_2": -1,
    "button_location": -1,
    "ai_signal_value": -1,
    "object_attributes": -1,
    "variable": -1,
    "timer": -1,
    "facet": -1,
    "play_sound": -1,
    "player_color": -1,
    "color_mood": -1,
    "reset_timer": -1,
    "object_state": -1,
    "action_type": -1,
    "message": "",
    "sound_name": "",
    "selected_object_ids": -1,
}

# Set using the version json files
effect_names = {}
default_attributes = {}
attributes = {}
attribute_presentation = {}



from __future__ import annotations

from enum import Enum
from typing import Union


class InfoDatasetBase(Enum):
    """
    This enum class is the base class for the UnitInfo, BuildingInfo, HeroInfo and OtherInfo datasets. It provides the
    properties and functions common to all the mentioned classes

    **Methods**
    
    >>> InfoDatasetBase.from_id()
    >>> InfoDatasetBase.from_dead_id()
    >>> InfoDatasetBase.from_icon_id()
    >>> InfoDatasetBase.from_hotkey_id()
    >>> InfoDatasetBase.gaia_only()
    >>> InfoDatasetBase.non_gaia()
    """

    @staticmethod
    def _id_map() -> dict:
        """
        **Private Method**

        Returns:
            A dict that maps the names of the properties to their index in the info tuple
        """
        return {'id': 0, 'icon_id': 1, 'dead_id': 2, 'hotkey_id': 3, 'gaia_only': 4}

    def _get_property(self, name: str) -> Union[int, bool]:
        """
        **Private Method**

        Get the specified property by its name from an object of this (or its derived) class

        Args:
            name: the name of the property to get

        Returns:
        """
        return self.value[self._id_map()[name]]

    @property
    def ID(self) -> int:
        """
        Returns:
            The ID of the specified unit
        """
        return self._get_property('id')

    @property
    def ICON_ID(self) -> int:
        """
        Returns:
            The Icon ID of the specified unit
        """
        return self._get_property('icon_id')

    @property
    def DEAD_ID(self) -> int:
        """
        Returns:
            The Dead Unit ID of the specified unit
        """
        return self._get_property('dead_id')

    @property
    def HOTKEY_ID(self) -> int:
        """
        Returns:
            The HotKey ID of the specified unit
        """
        return self._get_property('hotkey_id')

    @property
    def IS_GAIA_ONLY(self) -> bool:
        """
        Returns:
            A boolean value indicating if the specified unit is a gaia only unit (eg. Deer)
        """
        return self._get_property('gaia_only')

    @classmethod
    def _from_id(cls, id_type: str, value: int) -> InfoDatasetBase:
        """
        **Private Method**

        This function finds and returns the member object that uses the given value for the specified property (id_type)

        Args:
            id_type: the property of member objects that should match the value specified
            value: the value of the property to search for

        Returns:
            An InfoDatasetBase member object which uses the given value for the specified property (id_type
        """
        index = cls._id_map()[id_type]

        if type(value) is not int:
            raise TypeError(f"from_id expected int, got {type(value)}")
        if value < 0:
            raise ValueError(f"{value} is not a valid id value")

        for member in cls._member_map_.values():
            if member.value[index] == value:
                return member

        raise KeyError(f"A unit with {id_type} = {value} was not found in the dataset")

    @classmethod
    def from_id(cls, unit_id: int) -> InfoDatasetBase:
        """
        This function finds and returns the unit with the given unit ID

        Args:
            unit_id: the unit ID to search for

        Returns:
            A unit with the given unit I
        """
        return cls._from_id('id', unit_id)

    @classmethod
    def from_icon_id(cls, icon_id: int) -> InfoDatasetBase:
        """
        This function finds and returns the unit with the given icon ID

        Args:
            icon_id: the icon ID to search for

        Returns:
            A unit with the given icon ID
        """
        return cls._from_id('icon_id', icon_id)

    @classmethod
    def from_dead_id(cls, dead_id: int) -> InfoDatasetBase:
        """
        This function finds and returns the unit with the given dead unit ID

        Args:
            dead_id: the dead unit ID to search for

        Returns:
            A unit with the given dead unit ID
        """
        return cls._from_id('dead_id', dead_id)

    @classmethod
    def from_hotkey_id(cls, hotkey_id: int) -> InfoDatasetBase:
        """
        This function finds and returns the unit with the given hotkey ID. Note that there may be multiple units that
        use the same hotkey ID, currently only one is returned!

        Args:
            hotkey_id: the hotkey ID to search for

        Returns:
            A unit with the given hotkey ID
        """
        return cls._from_id('hotkey_id', hotkey_id)

    @classmethod
    def gaia_only(cls) -> list[InfoDatasetBase]:
        """
        Returns:
            A list of all the gaia only units (eg. Deer)
        """
        return cls._gaia_filter(gaia_only=True)

    @classmethod
    def non_gaia(cls) -> list[InfoDatasetBase]:
        """
        Returns:
            A list of all the units excluding gaia only units (eg. Deer)
        """
        return cls._gaia_filter(gaia_only=False)

    @classmethod
    def _gaia_filter(cls, gaia_only: bool) -> list[InfoDatasetBase]:
        """
        **Private Method**

        Args:
            gaia_only: if set to true, lists all gaia only units. If set to false, lists all units except gaia only units

        Returns:
            A list of either all gaia only units or a list of all units excluding gaia only units
        """
        result = []
        for x in cls:
            if x.IS_GAIA_ONLY == gaia_only:
                result.append(x)
        return result



from AoE2ScenarioParser.datasets.players import ColorId
from AoE2ScenarioParser.objects.aoe2_object import AoE2Object
from AoE2ScenarioParser.sections.retrievers.retriever_object_link import RetrieverObjectLink


class PlayerResources(AoE2Object):
    """Object for handling a tile in the map."""

    _link_list = [
        RetrieverObjectLink("food", "PlayerDataTwo", "resources[__index__].food"),
        RetrieverObjectLink("wood", "PlayerDataTwo", "resources[__index__].wood"),
        RetrieverObjectLink("gold", "PlayerDataTwo", "resources[__index__].gold"),
        RetrieverObjectLink("stone", "PlayerDataTwo", "resources[__index__].stone"),
        RetrieverObjectLink("color", "PlayerDataTwo", "resources[__index__].player_color"),
    ]

    def __init__(self, food: int, wood: int, gold: int, stone: int, color: int, **kwargs):
        super().__init__(**kwargs)

        self.food: int = food
        self.wood: int = wood
        self.gold: int = gold
        self.stone: int = stone
        self.color: ColorId = ColorId(color)



# This file was generated based on: versions/DE/v1.41/effects.json
from typing import Union

from AoE2ScenarioParser.datasets.effects import EffectId
from AoE2ScenarioParser.objects.data_objects.effect import Effect


class NewEffectSupport:
    def __init__(self, trigger_ref):
        self._trigger_ref = trigger_ref

    def none(
            self,
    ) -> Effect:
        return self._trigger_ref._add_effect(
            EffectId.NONE,
        )

    def change_diplomacy(
            self,
            diplomacy: Union[int, None] = None,
            source_player: Union[int, None] = None,
            target_player: Union[int, None] = None,
    ) -> Effect:
        return self._trigger_ref._add_effect(
            EffectId.CHANGE_DIPLOMACY,
            diplomacy=diplomacy,
            source_player=source_player,
            target_player=target_player,
        )

    def research_technology(
            self,
            source_player: Union[int, None] = None,
            technology: Union[int, None] = None,
            force_research_technology: Union[int, None] = None,
    ) -> Effect:
        return self._trigger_ref._add_effect(
            EffectId.RESEARCH_TECHNOLOGY,
            source_player=source_player,
            technology=technology,
            force_research_technology=force_research_technology,
        )

    def send_chat(
            self,
            source_player: Union[int, None] = None,
            string_id: Union[int, None] = None,
            message: Union[str, None] = None,
            sound_name: Union[str, None] = None,
    ) -> Effect:
        return self._trigger_ref._add_effect(
            EffectId.SEND_CHAT,
            source_player=source_player,
            string_id=string_id,
            message=message,
            sound_name=sound_name,
        )

    def play_sound(
            self,
            source_player: Union[int, None] = None,
            location_x: Union[int, None] = None,
            location_y: Union[int, None] = None,
            location_object_reference: Union[int, None] = None,
            sound_name: Union[str, None] = None,
    ) -> Effect:
        return self._trigger_ref._add_effect(
            EffectId.PLAY_SOUND,
            source_player=source_player,
            location_x=location_x,
            location_y=location_y,
            location_object_reference=location_object_reference,
            sound_name=sound_name,
        )

    def tribute(
            self,
            quantity: Union[int, None] = None,
            tribute_list: Union[int, None] = None,
            source_player: Union[int, None] = None,
            target_player: Union[int, None] = None,
    ) -> Effect:
        return self._trigger_ref._add_effect(
            EffectId.TRIBUTE,
            quantity=quantity,
            tribute_list=tribute_list,
            source_player=source_player,
            target_player=target_player,
        )

    def unlock_gate(
            self,
            selected_object_ids: Union[int, None] = None,
    ) -> Effect:
        return self._trigger_ref._add_effect(
            EffectId.UNLOCK_GATE,
            selected_object_ids=selected_object_ids,
        )

    def lock_gate(
            self,
            selected_object_ids: Union[int, None] = None,
    ) -> Effect:
        return self._trigger_ref._add_effect(
            EffectId.LOCK_GATE,
            selected_object_ids=selected_object_ids,
        )

    def activate_trigger(
            self,
            trigger_id: Union[int, None] = None,
    ) -> Effect:
        return self._trigger_ref._add_effect(
            EffectId.ACTIVATE_TRIGGER,
            trigger_id=trigger_id,
        )

    def deactivate_trigger(
            self,
            trigger_id: Union[int, None] = None,
    ) -> Effect:
        return self._trigger_ref._add_effect(
            EffectId.DEACTIVATE_TRIGGER,
            trigger_id=trigger_id,
        )

    def ai_script_goal(
            self,
            ai_script_goal: Union[int, None] = None,
    ) -> Effect:
        return self._trigger_ref._add_effect(
            EffectId.AI_SCRIPT_GOAL,
            ai_script_goal=ai_script_goal,
        )

    def create_object(
            self,
            object_list_unit_id: Union[int, None] = None,
            source_player: Union[int, None] = None,
            location_x: Union[int, None] = None,
            location_y: Union[int, None] = None,
            facet: Union[int, None] = None,
    ) -> Effect:
        return self._trigger_ref._add_effect(
            EffectId.CREATE_OBJECT,
            object_list_unit_id=object_list_unit_id,
            source_player=source_player,
            location_x=location_x,
            location_y=location_y,
            facet=facet,
        )

    def task_object(
            self,
            object_list_unit_id: Union[int, None] = None,
            source_player: Union[int, None] = None,
            location_x: Union[int, None] = None,
            location_y: Union[int, None] = None,
            location_object_reference: Union[int, None] = None,
            area_x1: Union[int, None] = None,
            area_y1: Union[int, None] = None,
            area_x2: Union[int, None] = None,
            area_y2: Union[int, None] = None,
            object_group: Union[int, None] = None,
            object_type: Union[int, None] = None,
            action_type: Union[int, None] = None,
            selected_object_ids: Union[int, None] = None,
    ) -> Effect:
        return self._trigger_ref._add_effect(
            EffectId.TASK_OBJECT,
            object_list_unit_id=object_list_unit_id,
            source_player=source_player,
            location_x=location_x,
            location_y=location_y,
            location_object_reference=location_object_reference,
            area_x1=area_x1,
            area_y1=area_y1,
            area_x2=area_x2,
            area_y2=area_y2,
            object_group=object_group,
            object_type=object_type,
            action_type=action_type,
            selected_object_ids=selected_object_ids,
        )

    def declare_victory(
            self,
            source_player: Union[int, None] = None,
            enabled: Union[int, None] = None,
    ) -> Effect:
        return self._trigger_ref._add_effect(
            EffectId.DECLARE_VICTORY,
            source_player=source_player,
            enabled=enabled,
        )

    def kill_object(
            self,
            object_list_unit_id: Union[int, None] = None,
            source_player: Union[int, None] = None,
            area_x1: Union[int, None] = None,
            area_y1: Union[int, None] = None,
            area_x2: Union[int, None] = None,
            area_y2: Union[int, None] = None,
            object_group: Union[int, None] = None,
            object_type: Union[int, None] = None,
            selected_object_ids: Union[int, None] = None,
    ) -> Effect:
        return self._trigger_ref._add_effect(
            EffectId.KILL_OBJECT,
            object_list_unit_id=object_list_unit_id,
            source_player=source_player,
            area_x1=area_x1,
            area_y1=area_y1,
            area_x2=area_x2,
            area_y2=area_y2,
            object_group=object_group,
            object_type=object_type,
            selected_object_ids=selected_object_ids,
        )

    def remove_object(
            self,
            object_list_unit_id: Union[int, None] = None,
            source_player: Union[int, None] = None,
            area_x1: Union[int, None] = None,
            area_y1: Union[int, None] = None,
            area_x2: Union[int, None] = None,
            area_y2: Union[int, None] = None,
            object_group: Union[int, None] = None,
            object_type: Union[int, None] = None,
            object_state: Union[int, None] = None,
            selected_object_ids: Union[int, None] = None,
    ) -> Effect:
        return self._trigger_ref._add_effect(
            EffectId.REMOVE_OBJECT,
            object_list_unit_id=object_list_unit_id,
            source_player=source_player,
            area_x1=area_x1,
            area_y1=area_y1,
            area_x2=area_x2,
            area_y2=area_y2,
            object_group=object_group,
            object_type=object_type,
            object_state=object_state,
            selected_object_ids=selected_object_ids,
        )

    def change_view(
            self,
            source_player: Union[int, None] = None,
            location_x: Union[int, None] = None,
            location_y: Union[int, None] = None,
            scroll: Union[int, None] = None,
    ) -> Effect:
        return self._trigger_ref._add_effect(
            EffectId.CHANGE_VIEW,
            source_player=source_player,
            location_x=location_x,
            location_y=location_y,
            scroll=scroll,
        )

    def unload(
            self,
            object_list_unit_id: Union[int, None] = None,
            source_player: Union[int, None] = None,
            location_x: Union[int, None] = None,
            location_y: Union[int, None] = None,
            location_object_reference: Union[int, None] = None,
            area_x1: Union[int, None] = None,
            area_y1: Union[int, None] = None,
            area_x2: Union[int, None] = None,
            area_y2: Union[int, None] = None,
            object_group: Union[int, None] = None,
            object_type: Union[int, None] = None,
            selected_object_ids: Union[int, None] = None,
    ) -> Effect:
        return self._trigger_ref._add_effect(
            EffectId.UNLOAD,
            object_list_unit_id=object_list_unit_id,
            source_player=source_player,
            location_x=location_x,
            location_y=location_y,
            location_object_reference=location_object_reference,
            area_x1=area_x1,
            area_y1=area_y1,
            area_x2=area_x2,
            area_y2=area_y2,
            object_group=object_group,
            object_type=object_type,
            selected_object_ids=selected_object_ids,
        )

    def change_ownership(
            self,
            object_list_unit_id: Union[int, None] = None,
            source_player: Union[int, None] = None,
            target_player: Union[int, None] = None,
            area_x1: Union[int, None] = None,
            area_y1: Union[int, None] = None,
            area_x2: Union[int, None] = None,
            area_y2: Union[int, None] = None,
            object_group: Union[int, None] = None,
            object_type: Union[int, None] = None,
            flash_object: Union[int, None] = None,
            selected_object_ids: Union[int, None] = None,
    ) -> Effect:
        return self._trigger_ref._add_effect(
            EffectId.CHANGE_OWNERSHIP,
            object_list_unit_id=object_list_unit_id,
            source_player=source_player,
            target_player=target_player,
            area_x1=area_x1,
            area_y1=area_y1,
            area_x2=area_x2,
            area_y2=area_y2,
            object_group=object_group,
            object_type=object_type,
            flash_object=flash_object,
            selected_object_ids=selected_object_ids,
        )

    def patrol(
            self,
            object_list_unit_id: Union[int, None] = None,
            source_player: Union[int, None] = None,
            location_x: Union[int, None] = None,
            location_y: Union[int, None] = None,
            area_x1: Union[int, None] = None,
            area_y1: Union[int, None] = None,
            area_x2: Union[int, None] = None,
            area_y2: Union[int, None] = None,
            object_group: Union[int, None] = None,
            object_type: Union[int, None] = None,
            selected_object_ids: Union[int, None] = None,
    ) -> Effect:
        return self._trigger_ref._add_effect(
            EffectId.PATROL,
            object_list_unit_id=object_list_unit_id,
            source_player=source_player,
            location_x=location_x,
            location_y=location_y,
            area_x1=area_x1,
            area_y1=area_y1,
            area_x2=area_x2,
            area_y2=area_y2,
            object_group=object_group,
            object_type=object_type,
            selected_object_ids=selected_object_ids,
        )

    def display_instructions(
            self,
            object_list_unit_id: Union[int, None] = None,
            source_player: Union[int, None] = None,
            string_id: Union[int, None] = None,
            display_time: Union[int, None] = None,
            instruction_panel_position: Union[int, None] = None,
            play_sound: Union[int, None] = None,
            message: Union[str, None] = None,
            sound_name: Union[str, None] = None,
    ) -> Effect:
        return self._trigger_ref._add_effect(
            EffectId.DISPLAY_INSTRUCTIONS,
            object_list_unit_id=object_list_unit_id,
            source_player=source_player,
            string_id=string_id,
            display_time=display_time,
            instruction_panel_position=instruction_panel_position,
            play_sound=play_sound,
            message=message,
            sound_name=sound_name,
        )

    def clear_instructions(
            self,
            instruction_panel_position: Union[int, None] = None,
    ) -> Effect:
        return self._trigger_ref._add_effect(
            EffectId.CLEAR_INSTRUCTIONS,
            instruction_panel_position=instruction_panel_position,
        )

    def freeze_object(
            self,
            object_list_unit_id: Union[int, None] = None,
            source_player: Union[int, None] = None,
            area_x1: Union[int, None] = None,
            area_y1: Union[int, None] = None,
            area_x2: Union[int, None] = None,
            area_y2: Union[int, None] = None,
            object_group: Union[int, None] = None,
            object_type: Union[int, None] = None,
            selected_object_ids: Union[int, None] = None,
    ) -> Effect:
        return self._trigger_ref._add_effect(
            EffectId.FREEZE_OBJECT,
            object_list_unit_id=object_list_unit_id,
            source_player=source_player,
            area_x1=area_x1,
            area_y1=area_y1,
            area_x2=area_x2,
            area_y2=area_y2,
            object_group=object_group,
            object_type=object_type,
            selected_object_ids=selected_object_ids,
        )

    def use_advanced_buttons(
            self,
    ) -> Effect:
        return self._trigger_ref._add_effect(
            EffectId.USE_ADVANCED_BUTTONS,
        )

    def damage_object(
            self,
            quantity: Union[int, None] = None,
            object_list_unit_id: Union[int, None] = None,
            source_player: Union[int, None] = None,
            area_x1: Union[int, None] = None,
            area_y1: Union[int, None] = None,
            area_x2: Union[int, None] = None,
            area_y2: Union[int, None] = None,
            object_group: Union[int, None] = None,
            object_type: Union[int, None] = None,
            selected_object_ids: Union[int, None] = None,
    ) -> Effect:
        return self._trigger_ref._add_effect(
            EffectId.DAMAGE_OBJECT,
            quantity=quantity,
            object_list_unit_id=object_list_unit_id,
            source_player=source_player,
            area_x1=area_x1,
            area_y1=area_y1,
            area_x2=area_x2,
            area_y2=area_y2,
            object_group=object_group,
            object_type=object_type,
            selected_object_ids=selected_object_ids,
        )

    def place_foundation(
            self,
            object_list_unit_id: Union[int, None] = None,
            source_player: Union[int, None] = None,
            location_x: Union[int, None] = None,
            location_y: Union[int, None] = None,
    ) -> Effect:
        return self._trigger_ref._add_effect(
            EffectId.PLACE_FOUNDATION,
            object_list_unit_id=object_list_unit_id,
            source_player=source_player,
            location_x=location_x,
            location_y=location_y,
        )

    def change_object_name(
            self,
            object_list_unit_id: Union[int, None] = None,
            source_player: Union[int, None] = None,
            string_id: Union[int, None] = None,
            area_x1: Union[int, None] = None,
            area_y1: Union[int, None] = None,
            area_x2: Union[int, None] = None,
            area_y2: Union[int, None] = None,
            message: Union[str, None] = None,
            selected_object_ids: Union[int, None] = None,
    ) -> Effect:
        return self._trigger_ref._add_effect(
            EffectId.CHANGE_OBJECT_NAME,
            object_list_unit_id=object_list_unit_id,
            source_player=source_player,
            string_id=string_id,
            area_x1=area_x1,
            area_y1=area_y1,
            area_x2=area_x2,
            area_y2=area_y2,
            message=message,
            selected_object_ids=selected_object_ids,
        )

    def change_object_hp(
            self,
            quantity: Union[int, None] = None,
            object_list_unit_id: Union[int, None] = None,
            source_player: Union[int, None] = None,
            area_x1: Union[int, None] = None,
            area_y1: Union[int, None] = None,
            area_x2: Union[int, None] = None,
            area_y2: Union[int, None] = None,
            object_group: Union[int, None] = None,
            object_type: Union[int, None] = None,
            operation: Union[int, None] = None,
            selected_object_ids: Union[int, None] = None,
    ) -> Effect:
        return self._trigger_ref._add_effect(
            EffectId.CHANGE_OBJECT_HP,
            quantity=quantity,
            object_list_unit_id=object_list_unit_id,
            source_player=source_player,
            area_x1=area_x1,
            area_y1=area_y1,
            area_x2=area_x2,
            area_y2=area_y2,
            object_group=object_group,
            object_type=object_type,
            operation=operation,
            selected_object_ids=selected_object_ids,
        )

    def change_object_attack(
            self,
            armour_attack_quantity: Union[int, None] = None,
            armour_attack_class: Union[int, None] = None,
            object_list_unit_id: Union[int, None] = None,
            source_player: Union[int, None] = None,
            area_x1: Union[int, None] = None,
            area_y1: Union[int, None] = None,
            area_x2: Union[int, None] = None,
            area_y2: Union[int, None] = None,
            object_group: Union[int, None] = None,
            object_type: Union[int, None] = None,
            operation: Union[int, None] = None,
            selected_object_ids: Union[int, None] = None,
    ) -> Effect:
        return self._trigger_ref._add_effect(
            EffectId.CHANGE_OBJECT_ATTACK,
            armour_attack_quantity=armour_attack_quantity,
            armour_attack_class=armour_attack_class,
            object_list_unit_id=object_list_unit_id,
            source_player=source_player,
            area_x1=area_x1,
            area_y1=area_y1,
            area_x2=area_x2,
            area_y2=area_y2,
            object_group=object_group,
            object_type=object_type,
            operation=operation,
            selected_object_ids=selected_object_ids,
        )

    def stop_object(
            self,
            object_list_unit_id: Union[int, None] = None,
            source_player: Union[int, None] = None,
            area_x1: Union[int, None] = None,
            area_y1: Union[int, None] = None,
            area_x2: Union[int, None] = None,
            area_y2: Union[int, None] = None,
            object_group: Union[int, None] = None,
            object_type: Union[int, None] = None,
            selected_object_ids: Union[int, None] = None,
    ) -> Effect:
        return self._trigger_ref._add_effect(
            EffectId.STOP_OBJECT,
            object_list_unit_id=object_list_unit_id,
            source_player=source_player,
            area_x1=area_x1,
            area_y1=area_y1,
            area_x2=area_x2,
            area_y2=area_y2,
            object_group=object_group,
            object_type=object_type,
            selected_object_ids=selected_object_ids,
        )

    def attack_move(
            self,
            object_list_unit_id: Union[int, None] = None,
            source_player: Union[int, None] = None,
            location_x: Union[int, None] = None,
            location_y: Union[int, None] = None,
            location_object_reference: Union[int, None] = None,
            area_x1: Union[int, None] = None,
            area_y1: Union[int, None] = None,
            area_x2: Union[int, None] = None,
            area_y2: Union[int, None] = None,
            object_group: Union[int, None] = None,
            object_type: Union[int, None] = None,
            selected_object_ids: Union[int, None] = None,
    ) -> Effect:
        return self._trigger_ref._add_effect(
            EffectId.ATTACK_MOVE,
            object_list_unit_id=object_list_unit_id,
            source_player=source_player,
            location_x=location_x,
            location_y=location_y,
            location_object_reference=location_object_reference,
            area_x1=area_x1,
            area_y1=area_y1,
            area_x2=area_x2,
            area_y2=area_y2,
            object_group=object_group,
            object_type=object_type,
            selected_object_ids=selected_object_ids,
        )

    def change_object_armor(
            self,
            armour_attack_quantity: Union[int, None] = None,
            armour_attack_class: Union[int, None] = None,
            object_list_unit_id: Union[int, None] = None,
            source_player: Union[int, None] = None,
            area_x1: Union[int, None] = None,
            area_y1: Union[int, None] = None,
            area_x2: Union[int, None] = None,
            area_y2: Union[int, None] = None,
            object_group: Union[int, None] = None,
            object_type: Union[int, None] = None,
            operation: Union[int, None] = None,
            selected_object_ids: Union[int, None] = None,
    ) -> Effect:
        return self._trigger_ref._add_effect(
            EffectId.CHANGE_OBJECT_ARMOR,
            armour_attack_quantity=armour_attack_quantity,
            armour_attack_class=armour_attack_class,
            object_list_unit_id=object_list_unit_id,
            source_player=source_player,
            area_x1=area_x1,
            area_y1=area_y1,
            area_x2=area_x2,
            area_y2=area_y2,
            object_group=object_group,
            object_type=object_type,
            operation=operation,
            selected_object_ids=selected_object_ids,
        )

    def change_object_range(
            self,
            quantity: Union[int, None] = None,
            object_list_unit_id: Union[int, None] = None,
            source_player: Union[int, None] = None,
            area_x1: Union[int, None] = None,
            area_y1: Union[int, None] = None,
            area_x2: Union[int, None] = None,
            area_y2: Union[int, None] = None,
            object_group: Union[int, None] = None,
            object_type: Union[int, None] = None,
            operation: Union[int, None] = None,
            selected_object_ids: Union[int, None] = None,
    ) -> Effect:
        return self._trigger_ref._add_effect(
            EffectId.CHANGE_OBJECT_RANGE,
            quantity=quantity,
            object_list_unit_id=object_list_unit_id,
            source_player=source_player,
            area_x1=area_x1,
            area_y1=area_y1,
            area_x2=area_x2,
            area_y2=area_y2,
            object_group=object_group,
            object_type=object_type,
            operation=operation,
            selected_object_ids=selected_object_ids,
        )

    def change_object_speed(
            self,
            quantity: Union[int, None] = None,
            object_list_unit_id: Union[int, None] = None,
            source_player: Union[int, None] = None,
            area_x1: Union[int, None] = None,
            area_y1: Union[int, None] = None,
            area_x2: Union[int, None] = None,
            area_y2: Union[int, None] = None,
            object_group: Union[int, None] = None,
            object_type: Union[int, None] = None,
            selected_object_ids: Union[int, None] = None,
    ) -> Effect:
        return self._trigger_ref._add_effect(
            EffectId.CHANGE_OBJECT_SPEED,
            quantity=quantity,
            object_list_unit_id=object_list_unit_id,
            source_player=source_player,
            area_x1=area_x1,
            area_y1=area_y1,
            area_x2=area_x2,
            area_y2=area_y2,
            object_group=object_group,
            object_type=object_type,
            selected_object_ids=selected_object_ids,
        )

    def heal_object(
            self,
            quantity: Union[int, None] = None,
            object_list_unit_id: Union[int, None] = None,
            source_player: Union[int, None] = None,
            area_x1: Union[int, None] = None,
            area_y1: Union[int, None] = None,
            area_x2: Union[int, None] = None,
            area_y2: Union[int, None] = None,
            object_group: Union[int, None] = None,
            object_type: Union[int, None] = None,
            selected_object_ids: Union[int, None] = None,
    ) -> Effect:
        return self._trigger_ref._add_effect(
            EffectId.HEAL_OBJECT,
            quantity=quantity,
            object_list_unit_id=object_list_unit_id,
            source_player=source_player,
            area_x1=area_x1,
            area_y1=area_y1,
            area_x2=area_x2,
            area_y2=area_y2,
            object_group=object_group,
            object_type=object_type,
            selected_object_ids=selected_object_ids,
        )

    def teleport_object(
            self,
            object_list_unit_id: Union[int, None] = None,
            source_player: Union[int, None] = None,
            location_x: Union[int, None] = None,
            location_y: Union[int, None] = None,
            area_x1: Union[int, None] = None,
            area_y1: Union[int, None] = None,
            area_x2: Union[int, None] = None,
            area_y2: Union[int, None] = None,
            object_group: Union[int, None] = None,
            object_type: Union[int, None] = None,
            selected_object_ids: Union[int, None] = None,
    ) -> Effect:
        return self._trigger_ref._add_effect(
            EffectId.TELEPORT_OBJECT,
            object_list_unit_id=object_list_unit_id,
            source_player=source_player,
            location_x=location_x,
            location_y=location_y,
            area_x1=area_x1,
            area_y1=area_y1,
            area_x2=area_x2,
            area_y2=area_y2,
            object_group=object_group,
            object_type=object_type,
            selected_object_ids=selected_object_ids,
        )

    def change_object_stance(
            self,
            object_list_unit_id: Union[int, None] = None,
            source_player: Union[int, None] = None,
            area_x1: Union[int, None] = None,
            area_y1: Union[int, None] = None,
            area_x2: Union[int, None] = None,
            area_y2: Union[int, None] = None,
            object_group: Union[int, None] = None,
            object_type: Union[int, None] = None,
            attack_stance: Union[int, None] = None,
            selected_object_ids: Union[int, None] = None,
    ) -> Effect:
        return self._trigger_ref._add_effect(
            EffectId.CHANGE_OBJECT_STANCE,
            object_list_unit_id=object_list_unit_id,
            source_player=source_player,
            area_x1=area_x1,
            area_y1=area_y1,
            area_x2=area_x2,
            area_y2=area_y2,
            object_group=object_group,
            object_type=object_type,
            attack_stance=attack_stance,
            selected_object_ids=selected_object_ids,
        )

    def display_timer(
            self,
            string_id: Union[int, None] = None,
            display_time: Union[int, None] = None,
            time_unit: Union[int, None] = None,
            timer: Union[int, None] = None,
            reset_timer: Union[int, None] = None,
            message: Union[str, None] = None,
    ) -> Effect:
        return self._trigger_ref._add_effect(
            EffectId.DISPLAY_TIMER,
            string_id=string_id,
            display_time=display_time,
            time_unit=time_unit,
            timer=timer,
            reset_timer=reset_timer,
            message=message,
        )

    def enable_disable_object(
            self,
            object_list_unit_id: Union[int, None] = None,
            source_player: Union[int, None] = None,
            enabled: Union[int, None] = None,
    ) -> Effect:
        return self._trigger_ref._add_effect(
            EffectId.ENABLE_DISABLE_OBJECT,
            object_list_unit_id=object_list_unit_id,
            source_player=source_player,
            enabled=enabled,
        )

    def enable_disable_technology(
            self,
            source_player: Union[int, None] = None,
            technology: Union[int, None] = None,
            enabled: Union[int, None] = None,
    ) -> Effect:
        return self._trigger_ref._add_effect(
            EffectId.ENABLE_DISABLE_TECHNOLOGY,
            source_player=source_player,
            technology=technology,
            enabled=enabled,
        )

    def change_object_cost(
            self,
            object_list_unit_id: Union[int, None] = None,
            source_player: Union[int, None] = None,
            food: Union[int, None] = None,
            wood: Union[int, None] = None,
            stone: Union[int, None] = None,
            gold: Union[int, None] = None,
    ) -> Effect:
        return self._trigger_ref._add_effect(
            EffectId.CHANGE_OBJECT_COST,
            object_list_unit_id=object_list_unit_id,
            source_player=source_player,
            food=food,
            wood=wood,
            stone=stone,
            gold=gold,
        )

    def set_player_visibility(
            self,
            source_player: Union[int, None] = None,
            target_player: Union[int, None] = None,
            visibility_state: Union[int, None] = None,
    ) -> Effect:
        return self._trigger_ref._add_effect(
            EffectId.SET_PLAYER_VISIBILITY,
            source_player=source_player,
            target_player=target_player,
            visibility_state=visibility_state,
        )

    def change_object_icon(
            self,
            object_list_unit_id: Union[int, None] = None,
            source_player: Union[int, None] = None,
            area_x1: Union[int, None] = None,
            area_y1: Union[int, None] = None,
            area_x2: Union[int, None] = None,
            area_y2: Union[int, None] = None,
            object_group: Union[int, None] = None,
            object_type: Union[int, None] = None,
            object_list_unit_id_2: Union[int, None] = None,
            selected_object_ids: Union[int, None] = None,
    ) -> Effect:
        return self._trigger_ref._add_effect(
            EffectId.CHANGE_OBJECT_ICON,
            object_list_unit_id=object_list_unit_id,
            source_player=source_player,
            area_x1=area_x1,
            area_y1=area_y1,
            area_x2=area_x2,
            area_y2=area_y2,
            object_group=object_group,
            object_type=object_type,
            object_list_unit_id_2=object_list_unit_id_2,
            selected_object_ids=selected_object_ids,
        )

    def replace_object(
            self,
            object_list_unit_id: Union[int, None] = None,
            source_player: Union[int, None] = None,
            target_player: Union[int, None] = None,
            area_x1: Union[int, None] = None,
            area_y1: Union[int, None] = None,
            area_x2: Union[int, None] = None,
            area_y2: Union[int, None] = None,
            object_group: Union[int, None] = None,
            object_type: Union[int, None] = None,
            object_list_unit_id_2: Union[int, None] = None,
            selected_object_ids: Union[int, None] = None,
    ) -> Effect:
        return self._trigger_ref._add_effect(
            EffectId.REPLACE_OBJECT,
            object_list_unit_id=object_list_unit_id,
            source_player=source_player,
            target_player=target_player,
            area_x1=area_x1,
            area_y1=area_y1,
            area_x2=area_x2,
            area_y2=area_y2,
            object_group=object_group,
            object_type=object_type,
            object_list_unit_id_2=object_list_unit_id_2,
            selected_object_ids=selected_object_ids,
        )

    def change_object_description(
            self,
            object_list_unit_id: Union[int, None] = None,
            source_player: Union[int, None] = None,
            string_id: Union[int, None] = None,
            message: Union[str, None] = None,
    ) -> Effect:
        return self._trigger_ref._add_effect(
            EffectId.CHANGE_OBJECT_DESCRIPTION,
            object_list_unit_id=object_list_unit_id,
            source_player=source_player,
            string_id=string_id,
            message=message,
        )

    def change_player_name(
            self,
            source_player: Union[int, None] = None,
            string_id: Union[int, None] = None,
            message: Union[str, None] = None,
    ) -> Effect:
        return self._trigger_ref._add_effect(
            EffectId.CHANGE_PLAYER_NAME,
            source_player=source_player,
            string_id=string_id,
            message=message,
        )

    def change_train_location(
            self,
            object_list_unit_id: Union[int, None] = None,
            source_player: Union[int, None] = None,
            object_list_unit_id_2: Union[int, None] = None,
            button_location: Union[int, None] = None,
    ) -> Effect:
        return self._trigger_ref._add_effect(
            EffectId.CHANGE_TRAIN_LOCATION,
            object_list_unit_id=object_list_unit_id,
            source_player=source_player,
            object_list_unit_id_2=object_list_unit_id_2,
            button_location=button_location,
        )

    def change_research_location(
            self,
            source_player: Union[int, None] = None,
            technology: Union[int, None] = None,
            object_list_unit_id_2: Union[int, None] = None,
            button_location: Union[int, None] = None,
    ) -> Effect:
        return self._trigger_ref._add_effect(
            EffectId.CHANGE_RESEARCH_LOCATION,
            source_player=source_player,
            technology=technology,
            object_list_unit_id_2=object_list_unit_id_2,
            button_location=button_location,
        )

    def change_civilization_name(
            self,
            source_player: Union[int, None] = None,
            string_id: Union[int, None] = None,
            message: Union[str, None] = None,
    ) -> Effect:
        return self._trigger_ref._add_effect(
            EffectId.CHANGE_CIVILIZATION_NAME,
            source_player=source_player,
            string_id=string_id,
            message=message,
        )

    def create_garrisoned_object(
            self,
            object_list_unit_id: Union[int, None] = None,
            source_player: Union[int, None] = None,
            area_x1: Union[int, None] = None,
            area_y1: Union[int, None] = None,
            area_x2: Union[int, None] = None,
            area_y2: Union[int, None] = None,
            object_list_unit_id_2: Union[int, None] = None,
            selected_object_ids: Union[int, None] = None,
    ) -> Effect:
        return self._trigger_ref._add_effect(
            EffectId.CREATE_GARRISONED_OBJECT,
            object_list_unit_id=object_list_unit_id,
            source_player=source_player,
            area_x1=area_x1,
            area_y1=area_y1,
            area_x2=area_x2,
            area_y2=area_y2,
            object_list_unit_id_2=object_list_unit_id_2,
            selected_object_ids=selected_object_ids,
        )

    def acknowledge_ai_signal(
            self,
            ai_signal_value: Union[int, None] = None,
    ) -> Effect:
        return self._trigger_ref._add_effect(
            EffectId.ACKNOWLEDGE_AI_SIGNAL,
            ai_signal_value=ai_signal_value,
        )

    def modify_attribute(
            self,
            quantity: Union[int, None] = None,
            armour_attack_quantity: Union[int, None] = None,
            armour_attack_class: Union[int, None] = None,
            object_list_unit_id: Union[int, None] = None,
            source_player: Union[int, None] = None,
            operation: Union[int, None] = None,
            object_attributes: Union[int, None] = None,
            message: Union[str, None] = None,
    ) -> Effect:
        """
        The parameters 'armour_attack_quantity' and 'armour_attack_class' are only used when object_attributes is Armor
        or Attack (8 or 9). Use, 'quantity' otherwise.
        """
        if (armour_attack_quantity is not None or armour_attack_class is not None) and quantity is not None:
            raise ValueError("Cannot use 'armour_attack' attributes together with the 'quantity' attribute.")

        return self._trigger_ref._add_effect(
            EffectId.MODIFY_ATTRIBUTE,
            quantity=quantity,
            armour_attack_quantity=armour_attack_quantity,
            armour_attack_class=armour_attack_class,
            object_list_unit_id=object_list_unit_id,
            source_player=source_player,
            operation=operation,
            object_attributes=object_attributes,
            message=message
        )

    def modify_resource(
            self,
            quantity: Union[int, None] = None,
            tribute_list: Union[int, None] = None,
            source_player: Union[int, None] = None,
            operation: Union[int, None] = None,
    ) -> Effect:
        return self._trigger_ref._add_effect(
            EffectId.MODIFY_RESOURCE,
            quantity=quantity,
            tribute_list=tribute_list,
            source_player=source_player,
            operation=operation,
        )

    def modify_resource_by_variable(
            self,
            tribute_list: Union[int, None] = None,
            source_player: Union[int, None] = None,
            operation: Union[int, None] = None,
            variable: Union[int, None] = None,
    ) -> Effect:
        return self._trigger_ref._add_effect(
            EffectId.MODIFY_RESOURCE_BY_VARIABLE,
            tribute_list=tribute_list,
            source_player=source_player,
            operation=operation,
            variable=variable,
        )

    def set_building_gather_point(
            self,
            object_list_unit_id: Union[int, None] = None,
            source_player: Union[int, None] = None,
            location_x: Union[int, None] = None,
            location_y: Union[int, None] = None,
            area_x1: Union[int, None] = None,
            area_y1: Union[int, None] = None,
            area_x2: Union[int, None] = None,
            area_y2: Union[int, None] = None,
            selected_object_ids: Union[int, None] = None,
    ) -> Effect:
        return self._trigger_ref._add_effect(
            EffectId.SET_BUILDING_GATHER_POINT,
            object_list_unit_id=object_list_unit_id,
            source_player=source_player,
            location_x=location_x,
            location_y=location_y,
            area_x1=area_x1,
            area_y1=area_y1,
            area_x2=area_x2,
            area_y2=area_y2,
            selected_object_ids=selected_object_ids,
        )

    def script_call(
            self,
            string_id: Union[int, None] = None,
            message: Union[str, None] = None,
    ) -> Effect:
        return self._trigger_ref._add_effect(
            EffectId.SCRIPT_CALL,
            string_id=string_id,
            message=message,
        )

    def change_variable(
            self,
            quantity: Union[int, None] = None,
            operation: Union[int, None] = None,
            variable: Union[int, None] = None,
            message: Union[str, None] = None,
    ) -> Effect:
        return self._trigger_ref._add_effect(
            EffectId.CHANGE_VARIABLE,
            quantity=quantity,
            operation=operation,
            variable=variable,
            message=message,
        )

    def clear_timer(
            self,
            timer: Union[int, None] = None,
    ) -> Effect:
        return self._trigger_ref._add_effect(
            EffectId.CLEAR_TIMER,
            timer=timer,
        )

    def change_object_player_color(
            self,
            object_list_unit_id: Union[int, None] = None,
            source_player: Union[int, None] = None,
            area_x1: Union[int, None] = None,
            area_y1: Union[int, None] = None,
            area_x2: Union[int, None] = None,
            area_y2: Union[int, None] = None,
            player_color: Union[int, None] = None,
            selected_object_ids: Union[int, None] = None,
    ) -> Effect:
        return self._trigger_ref._add_effect(
            EffectId.CHANGE_OBJECT_PLAYER_COLOR,
            object_list_unit_id=object_list_unit_id,
            source_player=source_player,
            area_x1=area_x1,
            area_y1=area_y1,
            area_x2=area_x2,
            area_y2=area_y2,
            player_color=player_color,
            selected_object_ids=selected_object_ids,
        )

    def change_object_civilization_name(
            self,
            string_id: Union[int, None] = None,
            area_x1: Union[int, None] = None,
            area_y1: Union[int, None] = None,
            area_x2: Union[int, None] = None,
            area_y2: Union[int, None] = None,
            message: Union[str, None] = None,
            selected_object_ids: Union[int, None] = None,
    ) -> Effect:
        return self._trigger_ref._add_effect(
            EffectId.CHANGE_OBJECT_CIVILIZATION_NAME,
            string_id=string_id,
            area_x1=area_x1,
            area_y1=area_y1,
            area_x2=area_x2,
            area_y2=area_y2,
            message=message,
            selected_object_ids=selected_object_ids,
        )

    def change_object_player_name(
            self,
            object_list_unit_id: Union[int, None] = None,
            source_player: Union[int, None] = None,
            string_id: Union[int, None] = None,
            area_x1: Union[int, None] = None,
            area_y1: Union[int, None] = None,
            area_x2: Union[int, None] = None,
            area_y2: Union[int, None] = None,
            message: Union[str, None] = None,
            selected_object_ids: Union[int, None] = None,
    ) -> Effect:
        return self._trigger_ref._add_effect(
            EffectId.CHANGE_OBJECT_PLAYER_NAME,
            object_list_unit_id=object_list_unit_id,
            source_player=source_player,
            string_id=string_id,
            area_x1=area_x1,
            area_y1=area_y1,
            area_x2=area_x2,
            area_y2=area_y2,
            message=message,
            selected_object_ids=selected_object_ids,
        )

    def disable_unit_targeting(
            self,
            object_list_unit_id: Union[int, None] = None,
            source_player: Union[int, None] = None,
            area_x1: Union[int, None] = None,
            area_y1: Union[int, None] = None,
            area_x2: Union[int, None] = None,
            area_y2: Union[int, None] = None,
            selected_object_ids: Union[int, None] = None,
    ) -> Effect:
        return self._trigger_ref._add_effect(
            EffectId.DISABLE_UNIT_TARGETING,
            object_list_unit_id=object_list_unit_id,
            source_player=source_player,
            area_x1=area_x1,
            area_y1=area_y1,
            area_x2=area_x2,
            area_y2=area_y2,
            selected_object_ids=selected_object_ids,
        )

    def enable_unit_targeting(
            self,
            object_list_unit_id: Union[int, None] = None,
            source_player: Union[int, None] = None,
            area_x1: Union[int, None] = None,
            area_y1: Union[int, None] = None,
            area_x2: Union[int, None] = None,
            area_y2: Union[int, None] = None,
            selected_object_ids: Union[int, None] = None,
    ) -> Effect:
        return self._trigger_ref._add_effect(
            EffectId.ENABLE_UNIT_TARGETING,
            object_list_unit_id=object_list_unit_id,
            source_player=source_player,
            area_x1=area_x1,
            area_y1=area_y1,
            area_x2=area_x2,
            area_y2=area_y2,
            selected_object_ids=selected_object_ids,
        )

    def change_technology_cost(
            self,
            source_player: Union[int, None] = None,
            technology: Union[int, None] = None,
            food: Union[int, None] = None,
            wood: Union[int, None] = None,
            stone: Union[int, None] = None,
            gold: Union[int, None] = None,
    ) -> Effect:
        return self._trigger_ref._add_effect(
            EffectId.CHANGE_TECHNOLOGY_COST,
            source_player=source_player,
            technology=technology,
            food=food,
            wood=wood,
            stone=stone,
            gold=gold,
        )

    def change_technology_research_time(
            self,
            quantity: Union[int, None] = None,
            source_player: Union[int, None] = None,
            technology: Union[int, None] = None,
    ) -> Effect:
        return self._trigger_ref._add_effect(
            EffectId.CHANGE_TECHNOLOGY_RESEARCH_TIME,
            quantity=quantity,
            source_player=source_player,
            technology=technology,
        )

    def change_technology_name(
            self,
            source_player: Union[int, None] = None,
            technology: Union[int, None] = None,
            string_id: Union[int, None] = None,
            message: Union[str, None] = None,
    ) -> Effect:
        return self._trigger_ref._add_effect(
            EffectId.CHANGE_TECHNOLOGY_NAME,
            source_player=source_player,
            technology=technology,
            string_id=string_id,
            message=message,
        )

    def change_technology_description(
            self,
            source_player: Union[int, None] = None,
            technology: Union[int, None] = None,
            string_id: Union[int, None] = None,
            message: Union[str, None] = None,
    ) -> Effect:
        return self._trigger_ref._add_effect(
            EffectId.CHANGE_TECHNOLOGY_DESCRIPTION,
            source_player=source_player,
            technology=technology,
            string_id=string_id,
            message=message,
        )

    def enable_technology_stacking(
            self,
            source_player: Union[int, None] = None,
            technology: Union[int, None] = None,
    ) -> Effect:
        return self._trigger_ref._add_effect(
            EffectId.ENABLE_TECHNOLOGY_STACKING,
            source_player=source_player,
            technology=technology,
        )

    def disable_technology_stacking(
            self,
            source_player: Union[int, None] = None,
            technology: Union[int, None] = None,
    ) -> Effect:
        return self._trigger_ref._add_effect(
            EffectId.DISABLE_TECHNOLOGY_STACKING,
            source_player=source_player,
            technology=technology,
        )

    def acknowledge_multiplayer_ai_signal(
            self,
            ai_signal_value: Union[int, None] = None,
    ) -> Effect:
        return self._trigger_ref._add_effect(
            EffectId.ACKNOWLEDGE_MULTIPLAYER_AI_SIGNAL,
            ai_signal_value=ai_signal_value,
        )

    def disable_object_selection(
            self,
            object_list_unit_id: Union[int, None] = None,
            source_player: Union[int, None] = None,
            area_x1: Union[int, None] = None,
            area_y1: Union[int, None] = None,
            area_x2: Union[int, None] = None,
            area_y2: Union[int, None] = None,
            selected_object_ids: Union[int, None] = None,
    ) -> Effect:
        return self._trigger_ref._add_effect(
            EffectId.DISABLE_OBJECT_SELECTION,
            object_list_unit_id=object_list_unit_id,
            source_player=source_player,
            area_x1=area_x1,
            area_y1=area_y1,
            area_x2=area_x2,
            area_y2=area_y2,
            selected_object_ids=selected_object_ids,
        )

    def enable_object_selection(
            self,
            object_list_unit_id: Union[int, None] = None,
            source_player: Union[int, None] = None,
            area_x1: Union[int, None] = None,
            area_y1: Union[int, None] = None,
            area_x2: Union[int, None] = None,
            area_y2: Union[int, None] = None,
            selected_object_ids: Union[int, None] = None,
    ) -> Effect:
        return self._trigger_ref._add_effect(
            EffectId.ENABLE_OBJECT_SELECTION,
            object_list_unit_id=object_list_unit_id,
            source_player=source_player,
            area_x1=area_x1,
            area_y1=area_y1,
            area_x2=area_x2,
            area_y2=area_y2,
            selected_object_ids=selected_object_ids,
        )

    def change_color_mood(
            self,
            quantity: Union[int, None] = None,
            color_mood: Union[int, None] = None):
        return self._trigger_ref._add_effect(
            EffectId.CHANGE_COLOR_MOOD,
            quantity=quantity,
            color_mood=color_mood
        )



class TriggerCELock:
    def __init__(self, lock_conditions=False, lock_effects=False, lock_condition_type=None, lock_effect_type=None,
                 lock_condition_ids=None, lock_effect_ids=None):
        """
        Object used to identify which conditions and effects should be locked from change.

        Args:
            lock_conditions (bool): Lock all conditions
            lock_effects (bool): Lock all effects
            lock_condition_type (List[int]): Lock certain condition types. Example: `ConditionId.OWN_OBJECTS`
            lock_effect_type (List[int]): Lock certain effect types. Example: `EffectId.CREATE_OBJECT`
            lock_condition_ids (List[int]): Lock certain conditions by their id
            lock_effect_ids (List[int]): Lock certain effects by their id
        """
        if lock_condition_type is None:
            lock_condition_type = []
        if lock_effect_type is None:
            lock_effect_type = []
        if lock_condition_ids is None:
            lock_condition_ids = []
        if lock_effect_ids is None:
            lock_effect_ids = []

        self.lock_conditions = lock_conditions
        self.lock_effects = lock_effects
        self.lock_condition_type = lock_condition_type
        self.lock_effect_type = lock_effect_type
        self.lock_condition_ids = lock_condition_ids
        self.lock_effect_ids = lock_effect_ids



"""
Don't hate the player, hate the g... name.
https://memeguy.com/photos/images/quick-maffs-288168.jpg
"""
import math
from typing import Union


def sign(a: Union[int, float], b: Union[int, float]) -> int:
    """
    Args:
        a (int|float): The first number
        b (int|float): The second number

    Returns:
        Returns 1 if a is higher than b, -1 if b is higher than a, and 0 when they're equal
    """
    return 0 if a == b else math.copysign(1, a - b)



from typing import List

from AoE2ScenarioParser.objects.data_objects.unit import Unit
from AoE2ScenarioParser.objects.managers.unit_manager import UnitManager


class UnitManagerDE(UnitManager):

    def __init__(self, units: List[List[Unit]], **kwargs):
        super().__init__(units, **kwargs)



from typing import List

import AoE2ScenarioParser.datasets.conditions as condition_dataset
import AoE2ScenarioParser.datasets.effects as effect_dataset
from AoE2ScenarioParser.datasets.conditions import ConditionId
from AoE2ScenarioParser.datasets.effects import EffectId
from AoE2ScenarioParser.helper.exceptions import UnsupportedAttributeError
from AoE2ScenarioParser.helper.helper import exclusive_if
from AoE2ScenarioParser.helper.list_functions import list_changed, update_order_array, hash_list
from AoE2ScenarioParser.helper.string_manipulations import add_tabs
from AoE2ScenarioParser.objects.aoe2_object import AoE2Object
from AoE2ScenarioParser.objects.data_objects.condition import Condition
from AoE2ScenarioParser.objects.data_objects.effect import Effect
from AoE2ScenarioParser.objects.support.new_condition import NewConditionSupport
from AoE2ScenarioParser.objects.support.new_effect import NewEffectSupport
from AoE2ScenarioParser.objects.support.uuid_list import UuidList
from AoE2ScenarioParser.scenarios.scenario_store import getters
from AoE2ScenarioParser.sections.retrievers.retriever_object_link import RetrieverObjectLink


class Trigger(AoE2Object):
    """Object for handling a trigger."""

    _link_list = [
        RetrieverObjectLink("name", "Triggers", "trigger_data[__index__].trigger_name"),
        RetrieverObjectLink("description", "Triggers", "trigger_data[__index__].trigger_description"),
        RetrieverObjectLink("description_stid", "Triggers", "trigger_data[__index__].description_string_table_id"),
        RetrieverObjectLink("display_as_objective", "Triggers", "trigger_data[__index__].display_as_objective"),
        RetrieverObjectLink("short_description", "Triggers", "trigger_data[__index__].short_description"),
        RetrieverObjectLink("short_description_stid", "Triggers",
                            "trigger_data[__index__].short_description_string_table_id"),
        RetrieverObjectLink("display_on_screen", "Triggers", "trigger_data[__index__].display_on_screen"),
        RetrieverObjectLink("description_order", "Triggers", "trigger_data[__index__].objective_description_order"),
        RetrieverObjectLink("enabled", "Triggers", "trigger_data[__index__].enabled"),
        RetrieverObjectLink("looping", "Triggers", "trigger_data[__index__].looping"),
        RetrieverObjectLink("header", "Triggers", "trigger_data[__index__].make_header"),
        RetrieverObjectLink("mute_objectives", "Triggers", "trigger_data[__index__].mute_objectives"),
        RetrieverObjectLink("conditions", "Triggers", "trigger_data[__index__].condition_data",
                            process_as_object=Condition),
        RetrieverObjectLink("condition_order", "Triggers", "trigger_data[__index__].condition_display_order_array"),
        RetrieverObjectLink("effects", "Triggers", "trigger_data[__index__].effect_data",
                            process_as_object=Effect),
        RetrieverObjectLink("effect_order", "Triggers", "trigger_data[__index__].effect_display_order_array"),
        RetrieverObjectLink("trigger_id", retrieve_instance_number=True),
    ]

    def __init__(self,
                 name: str,
                 description: str = "",
                 description_stid: int = -1,
                 display_as_objective: int = 0,
                 short_description: str = "",
                 short_description_stid: int = -1,
                 display_on_screen: int = 0,
                 description_order: int = 0,
                 enabled: int = 1,
                 looping: int = 0,
                 header: int = 0,
                 mute_objectives: int = 0,
                 conditions: List[Condition] = None,
                 condition_order: List[int] = None,
                 effects: List[Effect] = None,
                 effect_order: List[int] = None,
                 trigger_id: int = -1,
                 **kwargs
                 ):
        super().__init__(**kwargs)

        if conditions is None:
            conditions = []
        if condition_order is None:
            condition_order = []
        if effects is None:
            effects = []
        if effect_order is None:
            effect_order = []

        self.name: str = name
        self.description: str = description
        self.description_stid: int = description_stid
        self.display_as_objective: int = display_as_objective
        self.short_description: str = short_description
        self.short_description_stid: int = short_description_stid
        self.display_on_screen: int = display_on_screen
        self.description_order: int = description_order
        self.enabled: int = enabled
        self.looping: int = looping
        self.header: int = header
        self.mute_objectives: int = mute_objectives
        self._condition_hash = hash_list(conditions)
        self.conditions: List[Condition] = conditions
        self.condition_order: List[int] = condition_order
        self._effect_hash = hash_list(effects)
        self.effects: List[Effect] = effects
        self.effect_order: List[int] = effect_order
        self.trigger_id: int = trigger_id

        self.new_effect: NewEffectSupport
        self.new_condition: NewConditionSupport

        self._assign_new_ce_support()

    def __deepcopy__(self, memo):
        cls = self.__class__
        result = cls.__new__(cls)
        memo[id(self)] = result
        for k, v in self.__dict__.items():
            if k in ['new_effect', 'new_condition']:
                continue
            setattr(result, k, self._deepcopy_entry(k, v))

        result._assign_new_ce_support()
        return result

    def _assign_new_ce_support(self):
        """Assigns new `new_effect` and `new_condition` objects to this trigger"""
        self.new_effect = NewEffectSupport(self)
        self.new_condition = NewConditionSupport(self)

    @property
    def condition_order(self):
        if list_changed(self.conditions, self._condition_hash):
            update_order_array(self._condition_order, len(self.conditions))
            self._condition_hash = hash_list(self.conditions)
        return self._condition_order

    @condition_order.setter
    def condition_order(self, val):
        self._condition_order = val

    @property
    def effect_order(self):
        if list_changed(self.effects, self._effect_hash):
            update_order_array(self._effect_order, len(self.effects))
            self._effect_hash = hash_list(self.effects)
        return self._effect_order

    @effect_order.setter
    def effect_order(self, val):
        self._effect_order = val

    @property
    def conditions(self) -> List[Condition]:
        return self._conditions

    @conditions.setter
    def conditions(self, val: List[Condition]) -> None:
        self._conditions = UuidList(self._host_uuid, val)
        self.condition_order = list(range(0, len(val)))

    @property
    def effects(self) -> List[Effect]:
        return self._effects

    @effects.setter
    def effects(self, val: List[Effect]) -> None:
        self._effects = UuidList(self._host_uuid, val)
        self.effect_order = list(range(0, len(val)))

    def _add_effect(self, effect_type: EffectId, ai_script_goal=None, armour_attack_quantity=None,
                    armour_attack_class=None, quantity=None, tribute_list=None, diplomacy=None,
                    object_list_unit_id=None, source_player=None, target_player=None, technology=None, string_id=None,
                    display_time=None, trigger_id=None, location_x=None, location_y=None,
                    location_object_reference=None, area_x1=None, area_y1=None, area_x2=None, area_y2=None,
                    object_group=None, object_type=None, instruction_panel_position=None, attack_stance=None,
                    time_unit=None, enabled=None, food=None, wood=None, stone=None, gold=None, item_id=None,
                    flash_object=None, force_research_technology=None, visibility_state=None, scroll=None,
                    operation=None, object_list_unit_id_2=None, button_location=None, ai_signal_value=None,
                    object_attributes=None, variable=None, timer=None, facet=None, play_sound=None, message=None,
                    player_color=None, sound_name=None, selected_object_ids=None, color_mood=None, reset_timer=None,
                    object_state=None, action_type=None, ) -> Effect:
        """Used to add new effect to trigger. Please use trigger.new_effect.<effect_name> instead"""

        def get_default_effect_attributes(eff_type):
            """Gets the default effect attributes based on a certain effect type, with exception handling"""
            sv = getters.get_scenario_version(self._host_uuid)
            try:
                return effect_dataset.default_attributes[eff_type]
            except KeyError:
                effect = EffectId(eff_type)
                raise UnsupportedAttributeError(
                    f"The effect {effect.name} is not supported in scenario version {sv}"
                ) from None

        effect_defaults = get_default_effect_attributes(effect_type)
        effect_attr = {}
        for key, value in effect_defaults.items():
            effect_attr[key] = (locals()[key] if locals()[key] is not None else value)
        new_effect = Effect(**effect_attr, host_uuid=self._host_uuid)
        self.effects.append(new_effect)
        return new_effect

    def _add_condition(self, condition_type: ConditionId, quantity=None,
                       attribute=None, unit_object=None, next_object=None, object_list=None,
                       source_player=None, technology=None, timer=None, area_x1=None, area_y1=None, area_x2=None,
                       area_y2=None, object_group=None, object_type=None, ai_signal=None, inverted=None, variable=None,
                       comparison=None, target_player=None, unit_ai_action=None, xs_function=None, object_state=None
                       ) -> Condition:
        """Used to add new condition to trigger. Please use trigger.new_condition.<condition_name> instead"""

        def get_default_condition_attributes(cond_type):
            """Gets the default condition attributes based on a certain condition type, with exception handling"""
            sv = getters.get_scenario_version(self._host_uuid)
            try:
                return condition_dataset.default_attributes[cond_type]
            except KeyError:
                condition = ConditionId(cond_type)
                raise UnsupportedAttributeError(
                    f"The condition {condition.name} is not supported in scenario version {sv}"
                ) from None

        condition_defaults = get_default_condition_attributes(condition_type)
        condition_attr = {}
        for key, value in condition_defaults.items():
            condition_attr[key] = (locals()[key] if locals()[key] is not None else value)
        new_condition = Condition(**condition_attr, host_uuid=self._host_uuid)
        self.conditions.append(new_condition)
        return new_condition

    def get_effect(self, effect_index: int = None, display_index: int = None) -> Effect:
        if not exclusive_if(effect_index is not None, display_index is not None):
            raise ValueError(f"Please identify an effect using either effect_index or display_index.")

        if effect_index is None:
            effect_index = self.effect_order[display_index]

        return self.effects[effect_index]

    def get_condition(self, condition_index: int = None, display_index: int = None) -> Condition:
        if not exclusive_if(condition_index is not None, display_index is not None):
            raise ValueError(f"Please identify a condition using either condition_index or display_index.")

        if condition_index is None:
            condition_index = self.condition_order[display_index]

        return self.conditions[condition_index]

    def remove_effect(self, effect_index: int = None, display_index: int = None, effect: Effect = None) -> None:
        if not exclusive_if(effect_index is not None, display_index is not None, effect is not None):
            raise ValueError(f"Please identify an effect using either effect_index, display_index or effect.")

        if effect is not None:
            effect_index = self.effects.index(effect)
        if effect_index is None:
            effect_index = self.effect_order[display_index]

        del self.effects[effect_index]

    def remove_condition(self, condition_index: int = None, display_index: int = None, condition: Condition = None) \
            -> None:
        if not exclusive_if(condition_index is not None, display_index is not None, condition is not None):
            raise ValueError(f"Please identify a condition using either condition_index, display_index or condition.")

        if condition is not None:
            condition_index = self.conditions.index(condition)

        if condition_index is None:
            condition_index = self.condition_order[display_index]

        del self.conditions[condition_index]

    def get_content_as_string(self, include_trigger_definition=False) -> str:
        return_string = ""

        data_tba = {
            'enabled': self.enabled != 0,
            'looping': self.looping != 0
        }

        if self.description != "":
            data_tba['description'] = f"'{self.description}'"
        if self.description_stid != -1:
            data_tba['description_stid'] = self.description_stid
        if self.short_description != "":
            data_tba['short_description'] = f"'{self.short_description}'"
        if self.short_description_stid != -1:
            data_tba['short_description_stid'] = self.short_description_stid
        if self.display_as_objective != 0:
            data_tba['display_as_objective'] = (self.display_as_objective != 0)
        if self.display_on_screen != 0:
            data_tba['display_on_screen'] = (self.display_on_screen != 0)
        if self.description_order != 0:
            data_tba['description_order'] = self.description_order
        if self.header != 0:
            data_tba['header'] = (self.header != 0)
        if self.mute_objectives != 0:
            data_tba['mute_objectives'] = (self.mute_objectives != 0)

        for key, value in data_tba.items():
            return_string += f"{key}: {value}\n"

        if len(self.condition_order) > 0:
            return_string += "conditions:\n"
            for c_display_order, condition_id in enumerate(self.condition_order):
                condition = self.conditions[condition_id]

                return_string += f"\t{condition_dataset.condition_names[condition.condition_type]} " \
                                 f"[Index: {condition_id}, Display: {c_display_order}]:\n"
                return_string += add_tabs(condition.get_content_as_string(), 2)

        if len(self.effect_order) > 0:
            return_string += "effects:\n"
            for e_display_order, effect_id in enumerate(self.effect_order):
                effect = self.effects[effect_id]

                return_string += f"\t{effect_dataset.effect_names[effect.effect_type]}" \
                                 f" [Index: {effect_id}, Display: {e_display_order}]:\n"
                return_string += add_tabs(effect.get_content_as_string(), 2)

        if include_trigger_definition:
            return f"\"{self.name}\" [Index: {self.trigger_id}]\n" + add_tabs(return_string, 1)
        return return_string

    def __str__(self) -> str:
        return f"[Trigger] {self.get_content_as_string(include_trigger_definition=True)}"



from typing import List

from AoE2ScenarioParser.objects.aoe2_object import AoE2Object
from AoE2ScenarioParser.sections.retrievers.retriever_object_link import RetrieverObjectLink


class PlayerDiplomacy(AoE2Object):
    """Object for handling a tile in the map."""

    _link_list = [
        RetrieverObjectLink("diplomacy_stance", "Diplomacy", "per_player_diplomacy[__index__].stance_with_each_player"),
    ]

    def __init__(self, diplomacy_stance: List[int], **kwargs):
        super().__init__(**kwargs)

        self.diplomacy_stance: List[int] = diplomacy_stance
